'''
File input/output handlers for running L4C Science.
'''

import numpy as np
import h5py

class L4CStreamingInputDataset(object):
    '''
    A context that provides streaming access to an L4C inputs HDF5 file; i.e.,
    instead of reading in an entire data array at once, allows a data array
    to be indexed at each time step.

    Parameters
    ----------
    hdf_file_path : str
        File path to inputs HDF5 dataset
    constants : dict
        Dictionary of constant values
    bounds : dict
        (Optional) Dictionary of bounds for each parameter
    fill : dict
        (Optional) Dictionary of fill values, as two-element tuples, for each
        parameter; used when that parameter exceeds its (lower, upper) bounds
        and defaults to (NaN, NaN)
    '''
    def __init__(
            self, hdf_file_path: str, constants: dict, bounds: dict = None,
            fill: dict = None):
        self.bounds = bounds
        self.fill = fill
        self.constants = constants
        self.file_path = hdf_file_path
        if bounds is not None:
            assert hasattr(bounds, 'keys'), 'Argument "bounds" should be a dict type'
        if fill is not None:
            for key in fill.keys():
                assert len(fill[key]) == 2,\
                    'Tuples in "fill" dictionary should have 2 elements only'

    def __enter__(self):
        self.data = h5py.File(self.file_path, 'r')
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.data.close()

    def __getitem__(self, key):
        return self.data[key]

    def derive(self, t, field):
        '''
        Some fields cannot be read directly from the input dataset but must
        be calculated based on other fields.

        Parameters
        ----------
        t : int
            Current time step
        field : str
            Name of the input driver to derive

        Returns
        -------
        numpy.ndarray
        '''
        if field == 'apar':
            par = self.data['drivers/par'][t,...]
            fpar = self.data['drivers/fpar'][t,...]
            return np.multiply(par[:,np.newaxis].repeat(81, axis = 1), fpar)

        if field == 'ft':
            k = self.constants['tsurf_freeze-thaw_threshold_kelvin']
            return np.where(self.data['drivers/tsurf'][t,...] <= k, 0, 1)

    def index(self, t, *fields):
        '''
        Get the driver data at time t for one or more fields.

        Parameters
        ----------
        t : int
            Current time step
        *fields : str
            One or more field names
        '''
        drivers = []
        for f in fields:
            if f not in self.data['drivers'].keys():
                slice = self.derive(t, f) # Some fields need to be calculated
            else:
                slice = self.data['drivers/%s' % f][t,...]

            # Protect against out-of-bounds values
            if self.bounds is not None:
                if f in self.bounds.keys():
                    if not np.isnan(self.bounds[f][0]) and not np.isnan(self.bounds[f][1]):
                        # Get the fill value, defaulting to NaN
                        bounds = self.bounds[f]
                        if self.fill is not None:
                            lfill, ufill = self.fill[f]
                            slice = np.where(
                                np.logical_or(slice < bounds[0], slice > bounds[1]),
                                    np.where(slice < bounds[0], lfill,
                                        np.where(slice > bounds[1], ufill, slice)), slice)
                        else:
                            slice = np.where(
                                np.logical_or(slice < bounds[0], slice > bounds[1]),
                                np.nan, slice)
            # Propagate any single-valued fields at 9-km to the 1-km subgrid
            if slice.ndim == 1:
                slice = slice[:,np.newaxis].repeat(81, axis = 1)
            drivers.append(slice)
        return drivers



'''
Test cases related to the pyl4c.apps.calibration module.
'''

import os
import unittest
import numpy as np
import pyl4c
from pyl4c.apps.calibration import BPLUT, cbar
from pyl4c.data.fixtures import restore_bplut, BPLUT as V4_BPLUT

BPLUT_PATH = os.path.join(os.path.dirname(pyl4c.__file__), 'data/files/SMAP_BPLUT_2020-07-31.csv')
BPLUT_CSV_FOUND = os.path.exists(BPLUT_PATH)
TEMP_HDF5_PATH = os.path.join(os.path.dirname(pyl4c.__file__), 'temp.h5')

class CalibrationTestSuite(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        if os.path.exists(TEMP_HDF5_PATH):
            os.remove(TEMP_HDF5_PATH)

    @unittest.skipIf(not BPLUT_CSV_FOUND, 'No path to recent BPLUT file')
    def test_bplut_with_network_access(self):
        'Should create in-memory representation of BPLUT with BPLUT() class'
        bp0 = BPLUT(restore_bplut(BPLUT_PATH))
        bp1 = BPLUT(bp0.data, hdf5_path = TEMP_HDF5_PATH)
        bp2 = BPLUT(hdf5_path = TEMP_HDF5_PATH)
        self.assertEqual(bp0.labels, bp1.labels)
        self.assertEqual(bp0.labels, bp2.labels)
        bp0_vals = bp0['LUE'].round(2)
        bp1_vals = bp1['LUE'].round(2)
        bp2_vals = bp2['LUE'].round(2)
        self.assertTrue(np.equal(
            bp0_vals[~np.isnan(bp0_vals)].round(2),
            bp1_vals[~np.isnan(bp1_vals)].round(2)).all())
        self.assertTrue(np.equal(
            bp1_vals[~np.isnan(bp1_vals)].round(2),
            bp2_vals[~np.isnan(bp2_vals)].round(2)).all())

    def test_bplut(self):
        'Should create in-memory representation of BPLUT with BPLUT() class'
        bp = BPLUT(V4_BPLUT)
        self.assertEqual(len(bp.labels), 18)
        self.assertTrue(np.equal(bp['LUE'][:,1], 1.71))

    def test_bplut_flat(self):
        'Should retrieve parameters in order for a given PFT'
        bp = BPLUT(V4_BPLUT)
        self.assertEqual(bp.flat(1).size, 18)
        self.assertEqual(bp.flat(1)[0], 1.71)

    def test_bplut_update(self):
        'Should update the correct entries in the BPLUT'
        # Test that entries are equal or are NaN (which is not equal to NaN)
        equal_or_nan = lambda x, y: np.equal(x[~np.isnan(x)], y[~np.isnan(y)])
        bp = BPLUT(V4_BPLUT)
        bp.update(0, (-9999, -9999), ('LUE', 'smsf0'), flush = False)
        bp.update(7, (-9999, -9999), ('CUE', 'smrz1'), flush = False)
        self.assertTrue(np.equal(
            bp.data['LUE'][0,np.array((0,7))],
            np.array((-9999, 2.53))).all())
        self.assertTrue(equal_or_nan(
            bp.data['smsf'][:,np.array((0,7))],
            np.array([[-9999, -42], [np.nan, 41]], dtype = np.float32)).all())
        self.assertTrue(equal_or_nan(
            bp.data['smrz'][:,np.array((0,7))],
            np.array([[np.nan, -15], [np.nan, -9999]], dtype = np.float32)).all())

    def test_cbar(self):
        'Should calculate cbar correctly'
        np.random.seed(9)
        rh = np.random.sample(10).reshape((10, 1))
        kmult = np.arange(0, 1, 0.1).reshape((10, 1))
        self.assertEqual(
            cbar(rh, kmult, q_rh = 75, q_k = 50).round(3)[0], 0.437)
        self.assertEqual(
            cbar(rh, kmult, q_rh = 85, q_k = 50).round(3)[0], 0.541)
        self.assertEqual(
            cbar(rh, kmult, q_rh = 85, q_k = 10).round(3)[0], 2.123)


if __name__ == '__main__':
    unittest.main()



import datetime
import hashlib
import os
import unittest
import h5py
import numpy as np
import pyl4c
from scipy.linalg import solve_banded
from scipy.sparse import dia_matrix
from pyl4c import haversine, pft_dominant, equal_or_nan
from pyl4c.ease2 import ease2_from_wgs84, ease2_to_wgs84, ease2_coords, ease2_search_radius, translate_row_col_to_ease2
from pyl4c.science import arrhenius, bias_correction_parameters, rescale_smrz, vpd, ordinals365, climatology365
from pyl4c.spatial import ease2_coords_approx
from pyl4c.utils import MockL4CGranule, composite, get_ease2_coords, get_ease2_slice_idx, get_ease2_slice_offsets, get_pft_array, get_slice_idx_by_bbox, index, partition, partition_generator, subset, summarize, summarize_by_class
from pyl4c.data.fixtures import ANCILLARY_DATA_PATHS as PATHS, restore_bplut
from l4c import L4CForwardProcessPointTestSuite
from transpiration import TranspirationTestSuite

ANCILLARY_DATA_FOUND = os.path.exists(
    PATHS['smap_l4c_ancillary_data_file_path'])
BPLUT_PATH = os.path.join(
    os.path.dirname(pyl4c.__file__),
    'data/files/SMAP_BPLUT_2020-07-31.csv')
BPLUT_CSV_FOUND = os.path.exists(BPLUT_PATH)
BBOX_CONUS = [-124.5, 24.4, -66.7, 50.0]

class CoordinateTransformations(unittest.TestCase):
    def test_ease2_coords(self):
        'Integration: Should generate EASE-Grid 2.0 arrays of WGS84 coords.'
        x, y = ease2_coords('M09')
        self.assertEqual(x.size, 3856)
        self.assertEqual(y.size, 1624)
        self.assertEqual(x.min().round(2), -179.95)
        self.assertEqual(x.max().round(2), 179.95)
        self.assertEqual(y.min().round(2), -84.66)
        self.assertEqual(y.max().round(2), 84.66)
        x, y = ease2_coords('M01')
        self.assertEqual(x.size, 34704)
        self.assertEqual(y.size, 14616)
        self.assertEqual(x.min().round(2), -179.99)
        self.assertEqual(x.max().round(2), 179.99)
        self.assertEqual(y.min().round(2), -85.0)
        self.assertEqual(y.max().round(2), 85.0)

    def test_ease2_search_radius(self):
        'Should accurately identify pixel coordinates in a search radius'
        coords = ease2_search_radius((-83, 42), 1)
        self.assertTrue(np.equal(coords,
            np.array([
                [267, 1038], [267, 1039], [268, 1038], [268, 1039],
                [267, 1037], [268, 1037], [266, 1038], [266, 1039],
                [266, 1037]
            ])).all())

    def test_translate_row_col_to_ease2(self):
        'Should correctly translate row-column coords to EASE-Grid 2.0 coords'
        ul = translate_row_col_to_ease2((0,0))
        lr = translate_row_col_to_ease2((1624,3856))
        self.assertTrue(np.equal(
            np.array(ul).round(1),
            np.array([-17363026.4,   7310036.8])).all())
        self.assertTrue(np.equal(
            np.array(lr).round(1),
            np.array([17372034.5, -7319044.9])).all())

    def test_ease2_grid_centering(self):
        'Integration: EASE-Grid 2.0 transforms should target grid cell center'
        r, c = ease2_from_wgs84(ease2_to_wgs84((0, 0), 'M09'), 'M01')
        self.assertTrue(r == 4 and c == 4)

    def test_ease2_grid_location_from_wgs84(self):
        'Should accurately locate row-column pair from WGS84 coordinate pair'
        r, c = ease2_from_wgs84((0, 0))
        self.assertTrue(r == 812 and c == 1928)
        r, c = ease2_from_wgs84((0, 0), exact = True)
        self.assertTrue(r == 811.5 and c == 1927.5)


class Data(unittest.TestCase):
    @unittest.skipIf(not BPLUT_CSV_FOUND, 'No path to recent BPLUT file')
    def test_restore_bplut(self):
        'Should read in a BPLUT stored as a CSV file'
        bp = restore_bplut(BPLUT_PATH)
        self.assertTrue(equal_or_nan(
            bp['LUE'].round(2),
            np.array([
                [ np.nan, 1.17, 1.4 , 2.03, 1.33, 0.95, 1.64, 2.49, 2.49, np.nan]
            ], dtype = np.float32)).all())


class Default(unittest.TestCase):
    def test_haversine(self):
        'Should accurately calculate the great circle distance'
        p1, p2 = ((-83.748333, 42.281389), (-83.045833, 42.331389))
        dist = haversine(p1, p2, radius = 6371e3)
        self.assertEqual(round(dist, 2), 58036.75)

    def test_pft_dominant(self):
        'Should accurately report dominant PFT'
        np.random.seed(9)
        pft = np.random.randint(0, 10, 50).reshape((5,10))
        self.assertTrue(
            np.equal(pft_dominant(pft), np.array([6, 8, 8, 8, 8])).all())


class Science(unittest.TestCase):
    def test_ordinals365(self):
        'Should correctly calculate a 365-day day-of-year index'
        dates = [
            datetime.date(2000, 1, 1) + datetime.timedelta(days = d)
            for d in range(0, 366 + 365)
        ]
        ordinals = ordinals365(dates)
        self.assertEqual(len(ordinals), 366 + 365)
        self.assertEqual(max(ordinals), 365)
        self.assertEqual(min(ordinals), 1)

    def test_climatology365(self):
        'Should correctly calculate a 365-day climatology'
        np.random.seed(9)
        arr = 10 * np.random.sample(366 + (365 * 2)) # e.g., 3 years
        dates = [
            datetime.date(2000, 1, 1) + datetime.timedelta(days = d)
            for d in range(0, 366 + (365 * 2))
        ]
        clim = climatology365(arr, dates)
        self.assertEqual(clim.mean().round(3), 5.032) # Same as arr.mean()
        idx = np.random.randint(0, arr.size, 90)
        arr[idx] = np.nan
        clim2 = climatology365(arr, dates)
        self.assertEqual(clim2.mean().round(3), 5.008)

    def test_arrhenius(self):
        'Should accurately calculate Arrhenius response to temperature'
        arr = arrhenius(np.arange(0, 25, 5) + 273.15, beta0 = 273.15).round(3)
        self.assertTrue(np.equal(
            np.array([0.166, 0.296, 0.478, 0.712, 1]), arr).all())
        arr = arrhenius(np.arange(0, 25, 5) + 273.15, beta0 = 288).round(3)
        self.assertTrue(np.equal(
            np.array([0.15 , 0.277, 0.459, 0.699, 1]), arr).all())

    def test_bias_correction_parameters(self):
        'Should calculate correct bias correction parameters'
        np.random.seed(9)
        arr = np.random.randint(0, 30, 30).reshape((15,2))
        self.assertTrue(np.equal(
            bias_correction_parameters(arr).round(3),
            np.array([7.353, 0.781])).all())
        np.random.seed(42)
        arr = np.random.randint(0, 30, 30).reshape((15,2))
        self.assertTrue(np.equal(
            bias_correction_parameters(arr, npoly = 2).round(3),
            np.array([-1.834,  1.036,  0.])).all())

    def test_rescale_smrz(self):
        'Should correctly re-scale root-zone soil moisture'
        np.random.seed(7)
        A = np.random.randint(0, 100, 9).reshape((3,3))
        self.assertTrue(np.equal(
            rescale_smrz(A, smrz_min = np.ones((1,3)) * 25).round(1),
            np.array([
                [75.2, 88.7,  5. ],
                [88.2, 94.8,  5. ],
                [97.7, 82.7,  5. ]])).all())
        self.assertTrue(np.equal(
            rescale_smrz(A, smrz_min = np.ones((1,3)) * 23).round(1),
            np.array([
                [76.4, 89.1, 31.4],
                [88.6, 94.9,  5. ],
                [97.8, 83.4,  5. ]])).all())
        self.assertTrue(np.equal(
            rescale_smrz(A, smrz_min = np.ones((1,3)) * 14).round(1),
            np.array([
                [80.6, 90.5, 59. ],
                [90.2, 95.5, 55.2],
                [98. , 85.9,  5. ]])).all())

    def test_tridiagonal_solver(self):
        'NOT a test of pyl4c; a test that pyl4c dependencies work as expected'
        A = np.array([ # Example from GitHub user cbellei
            [10, 2, 0, 0], [3, 10, 4, 0], [0, 1, 7 ,5], [0, 0, 3, 4]
        ], dtype = np.float32)
        B = np.array([ # Example from scipy.linalg.solve_banded()
            [5, 2, -1, 0, 0], [1, 4, 2, -1, 0], [0, 1, 3, 2, -1],
            [0, 0, 1, 2, 2], [0, 0, 0, 1, 1]
        ], dtype = np.float32)
        a0 = np.array([3, 4, 5, 6])
        b0 = np.array([0, 1, 2, 2, 3])
        sol1 = np.linalg.solve(A, a0)
        sol2 = solve_banded((1, 1), np.flipud(dia_matrix(A).data), a0)
        self.assertTrue(np.equal(sol1.round(4), sol2.round(4)).all())
        sol1 = np.linalg.solve(B, b0)
        sol2 = solve_banded((1, 2), np.flipud(dia_matrix(B).data), b0)
        self.assertTrue(np.equal(sol1.round(4), sol2.round(4)).all())

    def test_vpd(self):
        'Sould correctly calculate VPD'
        self.assertEqual(round(vpd(0.1, 1000, 273.15), 2), 459.14)
        self.assertEqual(round(vpd(0.1, 1000, 278.15), 2), 720.41)
        self.assertEqual(round(vpd(0.2, 1000, 278.15), 2), 585.27)
        self.assertEqual(round(vpd(0.2, 2000, 278.15), 2), 298.57)

class Subsetting(unittest.TestCase):
    bbox = BBOX_CONUS
    x_coords = np.arange(-179.5, 180, 1)
    y_coords = np.arange(-89.5, 90, 1)

    @classmethod
    def setUpClass(cls):
        cls.approx_ease2_coords = ease2_coords_approx('M09', in_1d = True)
        # NOTE: There is potential circularity here; if get_ease2*() functions
        #   aren't working then we can't create a mock
        cls.hdf_9km = MockL4CGranule()
        # TECHNICALLY sets up an integration test, as we're using a feature
        #   in spatial module
        cls.hdf_9km_sorta = MockL4CGranule(
            coords = ease2_coords_approx('M09', in_1d = False))

    @classmethod
    def tearDownClass(cls):
        cls.hdf_9km.close()
        cls.hdf_9km_sorta.close()

    def test_index(self):
        'Should index an array given arbitrary row, column indices'
        arr = np.arange(0, 36).reshape((6, 6))
        idx = [(4, 2, 3), (4, 3, 2)]
        self.assertTrue(np.equal([28, 15, 20], index(arr, idx)).all())

    @unittest.skipIf(not ANCILLARY_DATA_FOUND, 'No path to ancillary data')
    def test_get_ease2_coords(self):
        'Should retrieve EASE-Grid 2.0 coordinate arrays'
        kwargs = {'in_1d': True}
        x1, y1 = get_ease2_coords(grid = 'M01', **kwargs)
        x9, y9 = get_ease2_coords(grid = 'M09', **kwargs)
        self.assertAlmostEqual(round(float(np.min(x1)), 5), -179.99481)
        self.assertAlmostEqual(round(float(np.max(x1)), 5), 179.99481)
        self.assertAlmostEqual(round(float(np.min(x9)), 5), -179.95332)
        self.assertAlmostEqual(round(float(np.max(x9)), 5), 179.95332)

    @unittest.skipIf(not ANCILLARY_DATA_FOUND, 'No path to ancillary data')
    def test_get_ease2_slice_etc(self):
        'Should produce EASE-Grid 2.0 slice indices or offsets'
        kwargs = {'subset_id': 'CONUS'}
        self.assertEqual(
            get_ease2_slice_idx(grid = 'M01', **kwargs),
            ((5350, 10922), (1699, 4289)))
        self.assertEqual(
            get_ease2_slice_idx(grid = 'M09', **kwargs),
            ((594, 1214), (189, 477)))
        self.assertEqual(
            get_ease2_slice_offsets(grid = 'M01', **kwargs), (5350, 1699))
        self.assertEqual(
            get_ease2_slice_offsets(grid = 'M09', **kwargs), (594, 189))

    @unittest.skipIf(not ANCILLARY_DATA_FOUND, 'No path to ancillary data')
    def test_get_pft_array(self):
        'Should produce the desired PFT array, subset or not'
        kwargs = {'subset_id': 'CONUS'}
        self.assertTrue(
            np.equal(get_pft_array(grid = 'M01').shape, (14616, 34704)).all())
        self.assertTrue(
            np.equal(get_pft_array(grid = 'M09').shape, (1624, 3856)).all())
        self.assertTrue(
            np.equal(
                get_pft_array(
                    grid = 'M01', **kwargs).shape, (2590, 5572)).all())
        self.assertTrue(
            np.equal(
                get_pft_array(
                    grid = 'M09', **kwargs).shape, (288, 620)).all())

    def test_get_slice_idx_by_bbox(self):
        'Should correctly determine slicing indices based on a bounding box'
        nw_bbox = [-20, 10, -10, 20]   # Northern and Western hemispheres
        sw_bbox = [-20, -20, -10, -10] # Southern and Western hemispheres
        ne_bbox = [10, 10, 20, 20]     # ...
        se_bbox = [10, -20, 20, -10]

        nw_x_idx, nw_y_idx = get_slice_idx_by_bbox(
            self.x_coords, self.y_coords, subset_bbox = nw_bbox)
        nwx0, nwx1 = nw_x_idx
        nwy0, nwy1 = nw_y_idx
        sw_x_idx, sw_y_idx = get_slice_idx_by_bbox(
            self.x_coords, self.y_coords, subset_bbox = sw_bbox)
        swx0, swx1 = sw_x_idx
        swy0, swy1 = sw_y_idx
        ne_x_idx, ne_y_idx = get_slice_idx_by_bbox(
            self.x_coords, self.y_coords, subset_bbox = ne_bbox)
        nex0, nex1 = ne_x_idx
        ney0, ney1 = ne_y_idx
        se_x_idx, se_y_idx = get_slice_idx_by_bbox(
            self.x_coords, self.y_coords, subset_bbox = se_bbox)
        sex0, sex1 = se_x_idx
        sey0, sey1 = se_y_idx

        self.assertTrue(np.all(np.equal(self.x_coords[nwx0:nwx1], np.arange(-19.5, -10, 1))))
        self.assertTrue(np.all(np.equal(self.x_coords[swx0:swx1], np.arange(-19.5, -10, 1))))
        self.assertTrue(np.all(np.equal(self.x_coords[nex0:nex1], np.arange(10.5, 20, 1))))
        self.assertTrue(np.all(np.equal(self.x_coords[sex0:sex1], np.arange(10.5, 20, 1))))
        self.assertTrue(np.all(np.equal(self.y_coords[nwy0:nwy1], np.arange(10.5, 20, 1))))
        self.assertTrue(np.all(np.equal(self.y_coords[swy0:swy1], np.arange(-19.5, -10, 1))))
        self.assertTrue(np.all(np.equal(self.y_coords[ney0:ney1], np.arange(10.5, 20, 1))))
        self.assertTrue(np.all(np.equal(self.y_coords[sey0:sey1], np.arange(-19.5, -10, 1))))

        # Now, in reverse!
        nw_x_idx, nw_y_idx = get_slice_idx_by_bbox(
            self.x_coords[::-1], self.y_coords[::-1], subset_bbox = nw_bbox)
        nwx0, nwx1 = nw_x_idx
        nwy0, nwy1 = nw_y_idx
        sw_x_idx, sw_y_idx = get_slice_idx_by_bbox(
            self.x_coords[::-1], self.y_coords[::-1], subset_bbox = sw_bbox)
        swx0, swx1 = sw_x_idx
        swy0, swy1 = sw_y_idx
        ne_x_idx, ne_y_idx = get_slice_idx_by_bbox(
            self.x_coords[::-1], self.y_coords[::-1], subset_bbox = ne_bbox)
        nex0, nex1 = ne_x_idx
        ney0, ney1 = ne_y_idx
        se_x_idx, se_y_idx = get_slice_idx_by_bbox(
            self.x_coords[::-1], self.y_coords[::-1], subset_bbox = se_bbox)
        sex0, sex1 = se_x_idx
        sey0, sey1 = se_y_idx

        self.assertTrue(np.all(np.equal(
            self.x_coords[::-1][nwx0:nwx1], np.arange(-10.5, -20, -1))))
        self.assertTrue(np.all(np.equal(
            self.x_coords[::-1][swx0:swx1], np.arange(-10.5, -20, -1))))
        self.assertTrue(np.all(np.equal(
            self.x_coords[::-1][nex0:nex1], np.arange(19.5, 10, -1))))
        self.assertTrue(np.all(np.equal(
            self.x_coords[::-1][sex0:sex1], np.arange(19.5, 10, -1))))
        self.assertTrue(np.all(np.equal(
            self.y_coords[::-1][nwy0:nwy1], np.arange(19.5, 10, -1))))
        self.assertTrue(np.all(np.equal(
            self.y_coords[::-1][swy0:swy1], np.arange(-10.5, -20, -1))))
        self.assertTrue(np.all(np.equal(
            self.y_coords[::-1][ney0:ney1], np.arange(19.5, 10, -1))))
        self.assertTrue(np.all(np.equal(
            self.y_coords[::-1][sey0:sey1], np.arange(-10.5, -20, -1))))

    def test_partition_etc(self):
        'Should partition an array of positive indices into n pieces'
        self.assertEqual(
        partition(np.arange(1e3), 1), [(0, 1e3 + 1)])
        self.assertEqual(
        partition(np.arange(1e3), 2), [(0, 500), (500, 1001)])
        self.assertEqual(
        partition(np.arange(1e3), 3), [(0, 333), (333, 666), (666, 1001)])
        parts = [(0, 333), (333, 666), (666, 1001)]
        for i, part in enumerate(partition_generator(1e3, 3)):
            with self.subTest(i = i):
                self.assertEqual(part, parts[i])

    @unittest.skipIf(not ANCILLARY_DATA_FOUND, 'No path to ancillary data')
    def test_subset(self):
        'Should subset an input HDF5 file on EASE-Grid 2.0 grid'
        arr, xoff, yoff = subset(
            self.__class__.hdf_9km, 'NEE/nee_mean', subset_id = 'CONUS')
        self.assertTrue(np.equal(arr.shape, (288, 620)).all())
        self.assertTrue(xoff, 594)
        self.assertTrue(yoff, 189)

        arr, xoff, yoff = subset(
            self.__class__.hdf_9km, 'NEE/nee_mean',
            subset_bbox = [-124.5, 24.4, -66.7, 50.0])
        self.assertTrue(np.equal(arr.shape, (288, 620)).all())
        self.assertTrue(xoff, 594)
        self.assertTrue(yoff, 189)

    def test_subset2(self):
        'Should subset an input HDF5 file on (approximate) EASE-Grid 2.0 grid'
        # NOTE: This test uses affine transforms to obtain EASE-Grid 2.0
        #   coordinates, rather than the ancillary dataset's coordinates;
        #   this allows the test to be run *without* access to ancillary data
        arr, xoff, yoff = subset(
            self.__class__.hdf_9km_sorta, 'NEE/nee_mean', subset_id = 'CONUS')
        self.assertTrue(np.equal(arr.shape, (288, 620)).all())
        self.assertTrue(xoff, 594)
        self.assertTrue(yoff, 189)

        arr, xoff, yoff = subset(
            self.__class__.hdf_9km_sorta, 'NEE/nee_mean',
            subset_bbox = [-124.5, 24.4, -66.7, 50.0])
        self.assertTrue(np.equal(arr.shape, (288, 620)).all())
        self.assertTrue(xoff, 594)
        self.assertTrue(yoff, 189)

        # Also test the passing of x_coords and y_coords explicitly
        x_coords, y_coords = self.__class__.approx_ease2_coords
        arr, xoff, yoff = subset(
            self.__class__.hdf_9km_sorta, 'NEE/nee_mean', subset_id = 'CONUS',
            x_coords = x_coords, y_coords = y_coords)
        self.assertTrue(np.equal(arr.shape, (288, 620)).all())
        self.assertTrue(xoff, 594)
        self.assertTrue(yoff, 189)


class Compositing(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(9)

    def test_composite(self):
        'Should composite 2 or more arrays together, in parallel or serial'
        checksum = hashlib.sha256()
        arrays = [
            np.random.normal(255, 50, 100).astype(np.int16).reshape((10, 10))
            for i in range(0, 10)
        ]
        result = composite(*arrays, reducer = 'mean')
        checksum.update(result)
        self.assertEqual(checksum.digest(), b"\xc9M\x18.\x16\xf0%7W\xb5\xb3\x1aH\x0bn\x87z'xM\x9b\xcf\xb5\x9a\xc0\xaf\xae\xc8\n\xeaf\x1a")

        checksum = hashlib.sha256()
        result = composite(*arrays, reducer = 'mean', processes = 2)
        checksum.update(result)
        self.assertEqual(checksum.digest(), b"\xc9M\x18.\x16\xf0%7W\xb5\xb3\x1aH\x0bn\x87z'xM\x9b\xcf\xb5\x9a\xc0\xaf\xae\xc8\n\xeaf\x1a")


class Summarization(unittest.TestCase):
    def test_summarize(self):
        'Should correctly summarize the values in a numeric array'
        a = np.arange(0, 25).reshape((5, 5))
        b = a.copy() # Copy and place NoData into the array
        b.ravel()[np.in1d(b, b[1,])] = -9999
        c = a.copy()
        c.ravel()[np.in1d(c, c[1,])] = 1234.0
        self.assertEqual(summarize(a, {
            'mean': lambda x: x.sum() / x.size
        })['mean'], a.mean())
        self.assertEqual(summarize(a, {'mean': np.mean})['mean'], a.mean())
        self.assertEqual(
            summarize(a, {'mean': np.mean}, scale = 3)['mean'],
            np.multiply(a, 3).mean())

        # NoData
        self.assertEqual(
            summarize(b, {'mean': np.nanmean})['mean'], 13.25)
        self.assertEqual(
            summarize(
                c, {'mean': np.nanmean}, nodata = 1234.0)['mean'], 13.25)

        # Multiple summaries
        self.assertEqual(len(summarize(a, {
            'median': np.median,
            'mean': np.mean,
            'std': np.std
        })), 3)

    def test_summarize_by_class(self):
        'Should correctl summarize'
        a = np.arange(0, 25).reshape((5, 5))
        labels = np.ones((5, 5))
        self.assertEqual(
            summarize_by_class(
                a, labels, {'mean': np.nanmean})[1]['mean'], a.mean())

        labels2 = labels.copy() # Copy and place 0.0 into the array
        labels2.ravel()[np.in1d(a, a[1,])] = 0
        self.assertEqual(
            summarize_by_class(
                a, labels2, {'mean': np.nanmean})[1]['mean'], 13.25)

        self.assertEqual(
            len(summarize_by_class(
                a, labels2, {'mean': np.nanmean}).keys()), 1)
        self.assertEqual(
            len(summarize_by_class(
                a, labels2, {'mean': np.nanmean}, ignore = None).keys()), 2)



if __name__ == '__main__':
    unittest.main()



'''
Tools for down-scaling L4C 9-km data to 1-km scale, using the PFT means.

    nested = NestedGrid(subset_id = "CONUS")
    arr = nested.pft_mask(8) # 1-km map of PFT 8
    with h5py.File("something.h5", "r") as hdf:
        nested.downscale_hdf5_by_pft(hdf, "GPP/gpp_pft%d_mean")

The above example assumes that you have official SPL4CMDL HDF5 granules from
NSIDC or EarthData Search. It's also possible to downscale netCDF4 granules
from NASA AppEEARS, including granules that represent spatial subsets:

    nc = netCDF4.Dataset('AppEEARS_granule.nc')

    # Read the EASE-Grid 2.0 coordinate arrays, get bounds
    xmin, xmax = min(nc['fakedim1'][:]), max(nc['fakedim1'][:])
    ymin, ymax = min(nc['fakedim0'][:]), max(nc['fakedim0'][:])

    AOI = (xmin, ymin, xmax, ymax)
    # It may be necessary to provide a "shape" argument to ensure that
    #   the land-cover/ PFT data are forced to the shape of the netCDF4 data
    nested = NestedGrid(
        subset_bbox = AOI, shape = nc.variables['SOC_soc_mean'][0].shape)
    result = nested.downscale_netcdf_by_pft(
        nc, field = 'SOC_soc_pft%d_mean', dtype = np.float32)

All of the SMAP Level 4 Carbon (L4C) flux and soil organic carbon (SOC) state
variables are computed on a global, 1-km equal-area grid. We spatially average
the data to 9-km to meet operational constraints on data storage. However,
within each 9-km cell, we save the mean values for each Plant Functional Type
(PFT); e.g., the mean SOC density at all the "Evergreen Needleleaf" pixels in
a 9-km cell (between 0 and 81 pixels) is recorded, along with the means of up
to eight other PFTs. This allows a reconstruction of finer-scale detail, using
the 1-km map of PFT and assigning the mean values within each 9-km cell. The
9-km and 1-km grids are designed to nest perfectly [1]. **Therefore, it is
possible to generate a down-scaled 1-km map of SOC density (and of RH, NEE,
and GPP).**

1. https://nsidc.org/ease/ease-grid-projection-gt
'''

import os
import tempfile
import h5py
import numpy as np
from affine import Affine
from osgeo import gdal
from scipy.ndimage import zoom
from cached_property import cached_property
from pyl4c.data.fixtures import EASE2_GRID_PARAMS
from pyl4c.ease2 import ease2_coords
from pyl4c.utils import get_pft_array
from pyl4c.spatial import ease2_to_geotiff
from pyl4c.lib.cli import CommandLineInterface, ProgressBar

class NestedGrid(object):
    '''
    Represents a nested L4C data structure, where each grid cell has a mean
    value for each sub-grid PFT class. This allows the reconstruction of a
    finer scale grid by applying the PFT mean values to the subgrid. The
    arguments `subset_id` and `subset_bbox` are mutually exclusive.

    Parameters
    ----------
    pft : tuple
        PFT values to use in downscaling
    shape : Sequence
        (Optional) The intended shape (at 9-km resolution) of the L4C data
        that will be downscaled. If not provided, should the shape of the
        data not match the PFT map, an error will be raised.
    subset_bbox : Sequence
        (Optional) An optional bounding box, `(xmin, ymin, xmax, ymax)`,
        specifying a spatial subset that will be downscaled
    '''
    def __init__(self, pft = range(1, 9), shape = None, subset_bbox = None):
        self._offsets = (0, 0)
        self._pft_codes = pft
        self._shp_1km = EASE2_GRID_PARAMS['M01']['shape']
        self._shp_9km = EASE2_GRID_PARAMS['M09']['shape']
        self._slice_idx_1km = None
        self._slice_idx_9km = None
        self._subset_bbox = subset_bbox
        self._ul_coords = (subset_bbox[0], subset_bbox[-1]) # Upper-left coordinates
        self._lr_coords = (subset_bbox[-2], subset_bbox[1]) # Lower-right coordinates
        # This is the "global" affine transformation
        self._transform_1km = Affine.from_gdal(*EASE2_GRID_PARAMS['M01']['geotransform'])
        self._transform_9km = Affine.from_gdal(*EASE2_GRID_PARAMS['M09']['geotransform'])
        # This is the "output" (1-km) affine transformation
        self._transform = self._transform_1km * self._transform_1km.scale(1)
        if subset_bbox is not None:
            # Need to calculate "local" transformation
            gt = list(self._transform_9km.to_gdal())
            gt[0] = self._ul_coords[0]
            gt[3] = self._ul_coords[1]
            transform_9km = Affine.from_gdal(*gt)
            transform_1km = transform_9km * transform_9km.scale(1/9)
            # Figure out row-column coordinates of upper-left corner and
            #   the 9-km extent
            x0, y0 = list(map(int, ~self._transform_9km * self._ul_coords))
            x1, y1 = list(map(int, ~transform_9km * self._lr_coords))
            self._shp_9km = (y1, x1)
            self._slice_idx_9km = [(y0, y0 + y1), (x0, x0 + x1)]
            # Same for the 1-km extent
            x2, y2 = list(map(int, ~self._transform_1km * self._ul_coords))
            x3, y3 = list(map(int, ~transform_1km * self._lr_coords))
            self._shp_1km = (y3, x3)
            # Check that this is the expected shape; because of coordinate
            #   transformations with different libraries, we may need to
            #   fudge things a big
            if shape is not None:
                if self._shp_9km != shape:
                    print(f'WARNING: Expected shape {str(shape)} did not match actual shape {str(self._shp_9km)}; using expected shape')
                    # Adjust the width and height at the bottom-right corner
                    deltas = np.array(shape) - np.array(self._shp_9km)
                    self._shp_9km = shape
                    self._shp_1km = tuple(np.array(self._shp_9km) * 9)
                    self._slice_idx_9km = [
                        (y0, y0 + y1 + deltas[0]),
                        (x0, x0 + x1 + deltas[1])
                    ]
            self._slice_idx_1km = (np.array(self._slice_idx_9km) * 9).tolist()
            # Update the output 1-km transformation
            self._transform = transform_1km

    @cached_property
    def pft(self):
        'The 1-km PFT map'
        return get_pft_array(
            'M01', slice_idx = self._slice_idx_1km).astype(np.uint8)[np.newaxis,...]

    @property
    def shape(self):
        'Returns the 1-km array shape'
        return self._shp_1km

    def _downscale(self, downscaled, scale = 1, dtype = np.float32, nodata = -9999):
        # Where the PFT map is in the valid range, return data, else NoData
        return np.where(
            np.logical_and(
                np.in1d(self.pft.ravel(), self._pft_codes),
                downscaled.ravel() != nodata),
            np.multiply(downscaled.ravel(), scale), nodata)\
            .reshape(self._shp_1km)\
            .astype(dtype)

    def pft_mask(self, p):
        '''
        An (M x N) mask for selecting pixels matching specified PFT class.

        Parameters
        ----------
        p : int
            The numeric code for the PFT of interest

        Returns
        -------
        numpy.ndarray
        '''
        return np.where(self.pft == p, 1, 0)

    def downscale_hdf5_by_pft(*args, **kwargs):
        '''
        DEPRECATED. Use `NestedGrid.downscale_hdf_by_pft()` instead.
        '''
        self.downscale_hdf_by_pft(*args, **kwargs)

    def downscale_hdf_by_pft(
            self, hdf, field, scale = 1, dtype = np.float32, nodata = -9999):
        '''
        Resamples 9-km L4C data to 1-km scale by repeating the spatial mean
        values for each PFT on the 1-km land-cover grid.

        Will subset the *arrays if they are not the expected shape of the
        subset at 9-km scale. This function assumes that the grouped dataset,
        `hdf`, is either an official SPL4CMDL HDF5 granule or a netCDF4 file
        generated by NASA AppEEARS.

        NOTE: Use scale = 1e6 if a total flux (e.g., total GPP) is required;
        `1e6` is the number of square meters in a 1-km L4C pixel.

        Parameters
        ----------
        hdf : h5py.File or netCDF4.Dataset
        field : str
            Template for PFT-mean field in the HDF5 granule, e.g.,
            "GPP/gpp_pft%d_mean"
        scale : int or float
        dtype : numpy.dtype
        nodata : int or float

        Returns
        -------
        numpy.ndarray
        '''
        if nodata < 0:
            assert dtype not in (np.uint0, np.uint8, np.uint16, np.uint32, np.uint64),\
                "Can't use an unsigned data type with a negative NoData value"
        opts = { # Options to zoom()
            'order': 0, 'zoom': 9, 'mode': 'grid-constant', 'grid_mode': True
        }
        if self._subset_bbox is not None:
            y_idx, x_idx = self._slice_idx_9km
            ymin, ymax = y_idx
            xmin, xmax = x_idx

        downscaled = np.zeros(self._shp_1km, dtype = dtype)
        for p in self._pft_codes:
            # Allow for the possibility of either netCDF4 or h5py datasets
            if hasattr(hdf, 'variables'):
                source = hdf.variables[field % p]
            else:
                source = hdf[field % p]
            # If needed, subset the 9-km arrays, then down-scale to 1-km
            if self._subset_bbox is None:
                arr = source[:]
            else:
                arr = source[ymin:ymax, xmin:xmax]
            # Resize, multiply by PFT mask, then add to output where != NoData
            downscaled = np.add(
                downscaled, np.multiply(self.pft_mask(p), zoom(arr, **opts)))
        return self._downscale(downscaled, scale, dtype, nodata)

    def downscale_netcdf_by_pft(
            self, hdf, field, scale = 1, dtype = np.float32, nodata = -9999):
        '''
        Resamples 9-km L4C data to 1-km scale by repeating the spatial mean
        values for each PFT on the 1-km land-cover grid.

        This function assumes that the grouped dataset, `hdf`, is a netCDF4
        dataset from a file granule generated by NASA AppEEARS. If the granule
        is a spatial subset, that subset matches the bounds defined by the
        `subset_bbox` to `NestedGrid`. It allows for the netCDF4 granule to
        represent multiple time steps; i.e., arrays can be of the shape
        `(T, M, N)` where `T` is one or more time steps.

        NOTE: Use scale = 1e6 if a total flux (e.g., total GPP) is required;
        `1e6` is the number of square meters in a 1-km L4C pixel.

        Parameters
        ----------
        hdf : netCDF4.Dataset
        field : str
            Template for PFT-mean field in the HDF5 granule, e.g.,
            "GPP/gpp_pft%d_mean"
        scale : int or float
        dtype : numpy.dtype
        nodata : int or float

        Returns
        -------
        numpy.ndarray
        '''
        if nodata < 0:
            assert dtype not in (np.uint0, np.uint8, np.uint16, np.uint32, np.uint64),\
                "Can't use an unsigned data type with a negative NoData value"
        opts = { # Options to zoom()
            'order': 0, 'zoom': 9, 'mode': 'grid-constant', 'grid_mode': True
        }
        # Get the shape of the input 9-km arrays
        shp = hdf.variables[field % self._pft_codes[0]].shape
        # In NASA AppEEARS, it is possible to request multiple dates of data,
        #   in which case the resulting netCDF4 granule has 3D arrays where
        #   the first is the time axis
        if len(shp) > 2:
            if len(shp) > 3:
                raise ValueError(f'Too many dimensions in "{field % self._pft_codes[0]}" array; expected 2 or 3')
            # Assumes the first axis is a time axis
            dates = shp[0]
        else:
            dates = 1

        result = np.zeros((dates, *self._shp_1km), dtype = dtype)
        with ProgressBar(dates, 'Downscaling...') as progress:
            for t in range(0, dates):
                downscaled = np.zeros(self._shp_1km, dtype = dtype)
                for p in self._pft_codes:
                    # Allow for the possibility of either netCDF4 or h5py datasets
                    if hasattr(hdf, 'variables'):
                        source = hdf.variables[field % p]
                    else:
                        source = hdf[field % p]
                    if dates > 1:
                        arr = source[t,:]
                    else:
                        arr = source[:]
                    # Resize, multiply by PFT mask, then add to output where != NoData
                    downscaled = np.add(
                        downscaled, np.multiply(self.pft_mask(p), zoom(arr, **opts)))
                result[t] = self._downscale(downscaled, scale, dtype, nodata)
                progress.update(t)
        return result


class CLI(CommandLineInterface):
    '''
    Command-line interface for running the downscaling procedure.

        python resample.py run <hdf5_granule> --field="GPP/gpp_pft%d_mean"
            --subset-id="CONUS"

    Parameters
    ----------
    output_path : str
        The output file path
    pft : list or tuple
        The PFT codes to use in down-scaling; should probably be `range(1, 9)`
        (Default)
    field : str
        A Python formatting string representing the SPL4CMDL HDF5 data field
        name to be down-scaled, e.g., `"SOC/soc_pft%d_mean"` (Default) where
        `"%d"` will be filled-in with the numeric PFT code
    subset_id : str
        The name of a well-known geographic subset, see:
        `pyl4c.data.fixtures.SUBSETS_BBOX`
    scale : int or float
        A number to multiply pixels values against, e.g., `1e6` (1,000 square
        kilometers) to convert (g C m-2 day-1) to (g C day-1)
    nodata : int or float
        The NoData value (Default: -9999)
    dtype : str
        The NumPy data type (Default: `"float32"`)
    verbose : bool
        True to print information about the progress
    '''

    def __init__(
            self, output_path, pft = range(1, 9),
            field = 'SOC/soc_pft%d_mean', subset_id = None, scale = 1,
            nodata = -9999, dtype = 'float32', verbose = True):
        self._dtype = self.lookup_dtype(dtype)
        self._field_tpl = field
        self._nodata = nodata
        self._output_path = output_path
        self._pft = pft
        self._scale = scale
        self._subset_id = subset_id
        self._verbose = verbose

    def run(self, hdf5_path, compress = True):
        '''
        Downscales an L4C variable from the given HDF5 granule.

        Parameters
        ----------
        hdf5_path : str
            File path to an HDF5 granule
        compress : bool
            True to compress the output file (Default: True)
        '''
        nested = NestedGrid(self._pft, self._subset_id)
        with h5py.File(hdf5_path, 'r') as hdf:
            arr = nested.downscale_hdf5_by_pft(
                hdf, self._field_tpl, self._scale)
            if self._verbose:
                print('Downscaling...')
        xoff, yoff = nested.offsets
        output_path = self._output_path
        if compress:
            tmp = tempfile.NamedTemporaryFile()
            output_path = tmp.name
        # Write initial file
        ease2_to_geotiff(arr, output_path, 'M01', xoff = xoff, yoff = yoff)
        if not compress:
            return # Done
        # Optionally, compress the output file
        opts = gdal.TranslateOptions(
            format = 'GTiff', creationOptions = ['COMPRESS=LZW'])
        # Note that "output_path" is the temporary file
        gdal.Translate(self._output_path, output_path, options = opts)


if __name__ == '__main__':
    import fire
    fire.Fire(CLI)



'''
Utilities related to the TCF code developed by Joe Glassy, Lucas Jones. Use:

    $ python pyl4c/lib/tcf.py run /anx_lagr3/laj/smap/natv72/prelaunch/land/Y2000/D*
        --output_dir="/home/user/"


TODO:

- Support for reading spatial subsets; a kind of reverse sparse row-column
    mapping, e.g.:

        np.repeat(np.arange(0, 3856).reshape((1, 3856)), 1624, axis = 0)

Ideas for an alternative sparse array subsetting that can be done without
storing a large ancillary array:

- Count distance, in elements, between valid cells in the raveled array;
  the number of stored distances increases with the sparseness of the
  array; 0 indicates the adjacent cell (raveled) is valid.
'''

import datetime
import glob
import os
import re
import h5py
import numpy as np
from itertools import product
from cached_property import cached_property
from pyl4c.data.fixtures import ANCILLARY_DATA_PATHS, HDF_PATHS, EASE2_GRID_PARAMS
from pyl4c.utils import summarize, summarize_by_class

FIELD_MAP = { # Mapping TCF field names to L4C Ops field names
    'Ctot': HDF_PATHS['SPL4CMDL']['4']['SOC'],
    'Ctot*': HDF_PATHS['SPL4CMDL']['4']['SOC*'],
    'gpp': HDF_PATHS['SPL4CMDL']['4']['GPP'],
    'gpp*': HDF_PATHS['SPL4CMDL']['4']['GPP*'],
    'nee': HDF_PATHS['SPL4CMDL']['4']['NEE'],
    'nee*': HDF_PATHS['SPL4CMDL']['4']['NEE*'],
    'rhtot': HDF_PATHS['SPL4CMDL']['4']['RH'],
    'rhtot*': HDF_PATHS['SPL4CMDL']['4']['RH*'],
    'Tmult': 'EC/tmult_mean',
    'Wmult': 'EC/wmult_mean',
    'Emult': 'EC/emult_mean'
}

TYPE_MAP = {
    'flt32': np.float32,
    'flt16': np.float16,
    'int16': np.int16,
    'uint16': np.uint16,
    'int8': np.int8,
    'uint8': np.uint8
}

REV_FIELD_MAP = dict([(v, k) for k, v in FIELD_MAP.items()])
REV_TYPE_MAP = dict([(v, k) for k, v in TYPE_MAP.items()])


class MetaSparseArray(type):
    '''
    NOTE: These are potentially expensive, but necessary, operations; they are
    exposed at the global scope because I don't want successive instances of
    `SparseArray()` to do this lookup, and allocate more memory, each time.
    '''
    @property
    def sparse_row_idx(cls):
        if getattr(cls, '_sparse_row_idx', None) is None:
            cls._sparse_row_idx = np.fromfile(ANCILLARY_DATA_PATHS['smap_l4c_9km_sparse_row_index'],
                dtype = np.uint16)

        return cls._sparse_row_idx

    @property
    def sparse_col_idx(cls):
        if getattr(cls, '_sparse_col_idx', None) is None:
            cls._sparse_col_idx = np.fromfile(ANCILLARY_DATA_PATHS['smap_l4c_9km_sparse_col_index'],
                dtype = np.uint16)

        return cls._sparse_col_idx


class SparseArray(object, metaclass = MetaSparseArray):
    '''
    Represents a "sparse" or "land" array such as that output from the
    TCF C codebase.

    Parameters
    ----------
    source : str or numpy.ndarray
        Either a file system path to a stored output array in "sparse" form
        (assumes that the filename incorporates the data type, e.g.,
        `foo.flt32` or `foo.uint8`) or an extant `numpy.ndarray` (either
        sparse or not).
    grid : str
        The EASE-Grid 2.0 designation, e.g., "M09" for a 9-km grid or "M01"
        for a 1-km grid
    dtype : type
        The NumPy data type used to represent numbers (Default: None); must
        be specified when reading a NumPy or HDF5 array
    field : str
        Name of the HDF5 field to read
    '''
    def __init__(self, source, grid, dtype = None, field = None):
        self._dtype = dtype
        self._grid = grid
        self._field = field
        self._nodata = None

        # Incorporate information from the filename
        if isinstance(source, str):
            self.filename = source
            self._deflated = True
            self._ftype = self.filename.split('.')[-1]
            # Attempt to read the array data
            assert os.path.exists(source), 'File not found: %s' % source
            assert self._ftype in TYPE_MAP.keys(), 'Unrecognized file extension; could not determine data type'
            self.data = np.fromfile(source, dtype = TYPE_MAP[self._ftype])
        elif isinstance(source, np.ndarray):
            self.filename = None
            self._deflated = source.ndim == 1
            self._ftype = None
            assert self._dtype is not None,\
                'Must indicate the data type when reading an array!'
            assert source.ndim <= 2, 'No support for more than 2 axes'
            self.data = source
        elif isinstance(source, h5py.File):
            # NOTE: Trying to deflate the HDF5 array while reading it is
            #   inefficient because it requires looping over multiple index
            #   operations; it will never be faster than deflating the entire
            #   array one time
            assert self._field is not None, "__init__() missing 1 required positional argument: 'field'"
            self.filename = source.filename
            self._ftype = None
            self.data = source[self._field][:]
            self._deflated = self.data.ndim == 1
            assert self._dtype is not None,\
                'Must indicate the data type when reading an array!'
        else:
            raise TypeError('Must provide either a binary file path or a numpy.ndarray instance')

    def __nested_src_idx__(self, x, y, ny):
        # Serialization offset for nested grid elements, e.g., 0 through 8
        #   for a 1-km grid nested in a 9-km grid
        return (x * ny) + y

    def __nested_dst_idx__(self, x, y, m, n, nn, ny):
        # Serialization offset for a grid nested within a 9-km grid
        # x, y correspond to 9-km grid indices; m, n to nested grid indices;
        #   nn is number of nested columns (e.g., 9 for 1-km grid) and ny
        #   is the number of columns in the overall 2D grid (e.g., 34704)
        return (((x * nn) + m) * ny) + (y * nn) + n

    def __inflate_1km_grid__(self, nodata = -9999):
        # This is SLOW; it shouldn't be used in production, but just in case
        #   a pure Python version of mkgrid for 1-km grids is ever needed...
        shp = EASE2_GRID_PARAMS[self._grid]['shape']
        # NOTE: 1-km destination array starts raveled
        result = np.ones((shp[0] * shp[1],)) * nodata
        # Source array has first 81 pixels, then second 81 pixels, etc...
        nested_grid = product(range(0, 9), range(0, 9))
        for i in range(0, self.__class__.sparse_row_idx.shape[0]):
            row = self.__class__.sparse_row_idx[i]
            col = self.__class__.sparse_col_idx[i]
            for m, n in nested_grid:
                # This is a translation of the function from spland.c
                #   in the l4c-utils/mkgrid C library
                src = self.__nested_src_idx__(
                    i, self.__nested_src_idx__(m, n, 9), 81)
                dst = self.__nested_dst_idx__(row, col, m, n, 9, shp[1])
                result[dst] = self.data[src]

    @property
    def shape(self):
        "Alias for the array data's shape"
        return self.data.shape

    @classmethod
    def get_deflated_idx(cls, coords):
        '''
        Translates (row, column) coordinates in the inflated array to the
        corresponding 1D index in the deflated (sparse) array.

        Parameters
        ----------
        coords : tuple or list
            Sequence of (row, column) integer indices

        Returns
        -------
        numpy.ndarray
        '''
        row, col = coords
        return np.arange(0, cls.sparse_row_idx.shape[0])[
            np.logical_and(cls.sparse_col_idx == col, cls.sparse_row_idx == row)
        ]

    @classmethod
    def sparse_selector(cls, slice_idx, grid = 'M09'):
        '''
        Returns a 1D array with values in `[np.nan, 1]` where 1 is assigned to
        those cells inside the desired spatial subset of the corresponding
        2D array. This allows efficient summarization, for instance, of the
        sparse (1D) data array without having to inflate it.

        Parameters
        ----------
        slice_idx : tuple or list
            Sequence of nested bounding box coordinates in pixel space, e.g.:
            `((xmin, xmax), (ymin, ymax))`
        grid : str
            EASE-Grid 2.0 designation describing the size of the grid that
            should be restored, e.g., "M09" or "M01"

        Returns
        -------
        numpy.ndarray
        '''
        shp = EASE2_GRID_PARAMS[grid]['shape']
        x_idx, y_idx = slice_idx
        xmin, xmax = x_idx
        ymin, ymax = y_idx

        # If the array is already deflated, we need to generate a
        #   corresponding 1D array to "select" values
        # Get a grid of NaNs, then fill in 1 where the subset fits
        selector_deflated = np.ones(cls.sparse_row_idx.shape) * np.nan
        selector = np.ones(shp) * np.nan
        selector[ymin:ymax, xmin:xmax] = 1
        # Exactly like self.inflate() but selector is the "data'"
        for i in range(0, cls.sparse_row_idx.shape[0]):
            row = cls.sparse_row_idx[i]
            col = cls.sparse_col_idx[i]
            selector_deflated[i] = selector[row, col]

        return selector_deflated

    @classmethod
    def ordinal_to_date(cls, year, ordinal):
        '''
        Returns a `datetime.date()` instance based on year, ordinal day.

        Parameters
        ----------
        year : int
        ordinal : int
            Day of the year on interval [1, 366]

        Returns
        -------
        datetime.date
        '''
        assert ordinal <= 366, 'Ordinal out of range [1, 366]'
        assert year in range(1800, 2100), 'Year out of range [1800, 2100]'
        # In leap years, for days after February 29, subtract 1 so that
        #   we obtain the "regular" date
        if year % 4 == 0:
            if ordinal == 60:
                return datetime.date(year, 2, 29)
                if ordinal > 60:
                    ordinal -= 1
                    d = datetime.date.fromordinal(ordinal)
                    return datetime.date(year, d.month, d.day)

    def deflate(self, nodata = -9999):
        '''
        Inverse of `SparseArray.inflate()`; converts a 2D array to a "sparse"
        or "land" representation.

        Parameters
        ----------
        nodata : int or float
            The NoData value to output in resulting array
        '''
        # If the array is already deflated but a different NoData value was
        #   requested, update that value
        if self._deflated:
            if self._nodata is not None and nodata != self._nodata:
                self.data = np.where(
                    self.data == self._nodata, nodata, self.data)
            else:
                return None # Otherwise, return nothing

        # Initialize an empty array (fill with NoData)
        result = np.ones(
            self.__class__.sparse_row_idx.shape, self._dtype) * nodata
        for i in range(0, self.__class__.sparse_row_idx.shape[0]):
            row = self.__class__.sparse_row_idx[i]
            col = self.__class__.sparse_col_idx[i]
            result[i] = self.data[row, col]
        self._deflated = True
        self.data = result
        result = None

    def inflate(self, nodata = -9999):
        '''
        Converts from a "sparse" or "land" array to an EASE-Grid 2.0 array. The
        sparse format was developed by Joe Glassy and Lucas Jones. Special files
        are needed to index the sparse data correctly (it is not simply a raveled
        array filtered to land pixels).

        Parameters
        ----------
        nodata : int or float
            The NoData value to output in resulting array
        '''
        # If the array is already inflated but a different NoData value was
        #   requested, update that value
        if not self._deflated:
            if self._nodata is not None and nodata != self._nodata:
                self.data = np.where(
                    self.data == self._nodata, nodata, self.data)
            else:
                return None # Otherwise, return nothing

        # Initialize an empty array (fill with NoData)
        shp = EASE2_GRID_PARAMS[self._grid]['shape']
        result = np.ones(shp, self._dtype) * nodata
        for i in range(0, self.__class__.sparse_row_idx.shape[0]):
            row = self.__class__.sparse_row_idx[i]
            col = self.__class__.sparse_col_idx[i]
            result[row, col] = self.data[i]
        self._deflated = False
        self.data = result
        result = None

    def inflate_subset(self, slice_idx, nodata = -9999):
        '''
        Returns a subset view of the inflated array, for convenience. This is
        not faster than subsetting the inflated array.

        Parameters
        ----------
        slice_idx : tuple or list
            `((xmin, xmax), (ymin, ymax))` in pixel space
        nodata : int or float
            The NoData value to output in resulting array

        Returns
        -------
        numpy.ndarray
        '''
        x_idx, y_idx = slice_idx
        if not self._deflated:
            return self.data[y_idx[0]:y_idx[1], x_idx[0]:x_idx[1]]

        selector = self.__class__.sparse_selector(slice_idx, self._grid)
        xs = x_idx[1] - x_idx[0] # Get the output array size
        ys = y_idx[1] - y_idx[0]
        shp = (ys, xs)
        data = np.multiply(self.data, selector)
        # Initialize an empty array (fill with NoData)
        result = np.ones(shp) * nodata
        for i in range(0, self.__class__.sparse_row_idx.shape[0]):
            row = self.__class__.sparse_row_idx[i]
            col = self.__class__.sparse_col_idx[i]
            if row < y_idx[1] and col < x_idx[1]:
                row -= y_idx[0]
                col -= x_idx[0]
                if row >= 0 and col >= 0:
                    result[row, col] = data[i]
        return result

    def replace(self, old, new):
        '''
        Replace "old" values with "new" values, e.g., replace old NoData
        values with a new value.

        Parameters
        ----------
        old : int or float
            Value to be replaced
        new : int or float
            Value to replace the old value
        '''
        self.data[self.data == old] = new

    def summarize(self, selector = None, **kwargs):
        '''
        Summarize the values of an array (inflated or deflated) without
        inflating it, optionally for a spatial subset.

        Parameters
        ----------
        selector : numpy.ndarray
            An array that describes which which values should be summarized;
            all non-NaN entries correspond to entries in the data array that
            will be summarized.
        **kwargs
            Additional arguments to `pyl4c.utils.summarize()` or
            `pyl4c.utils.summarize_by_class()`

        Returns
        -------
        dict
        '''
        shp = EASE2_GRID_PARAMS[self._grid]['shape']
        ymin = xmin = 0 # Default is full extent of grid
        ymax, xmax = shp
        if not self._deflated:
            return summarize(self.data[ymin:ymax, xmin:xmax], **kwargs)
        if selector is None:
            selector = np.ones(self.shape)
        # Selector multiplies NaN against areas outside the subset
        return summarize(np.multiply(self.data, selector), **kwargs)

    def summarize_by_class(
            self, selector = None, class_array = None, **kwargs):
        '''
        Summarize the values of an array (inflated or deflated) without
        inflating it, optionally for a spatial subset.

        Parameters
        ----------
        selector : numpy.ndarray
            An array that describes which which values should be summarized;
            all non-NaN entries correspond to entries in the data array that
            will be summarize
        class_array : numpy.ndarray
            An array that has discrete class labels (e.g., decimal numbers);
            values that share a class will be aggregated in the summary
        **kwargs
            Additional arguments to `pyl4c.utils.summarize()` or
            `pyl4c.utils.summarize_by_class()`

        Returns
        -------
        dict
        '''
        assert class_array.shape == self.shape, 'class_array does not have the same shape!'
        shp = EASE2_GRID_PARAMS[self._grid]['shape']
        ymin = xmin = 0 # Default is full extent of grid
        ymax, xmax = shp
        if not self._deflated:
            return summarize_by_class(
                self.data[ymin:ymax, xmin:xmax], class_array, **kwargs)
        if selector is None:
            selector = np.ones(self.shape)
        # Selector multiplies NaN against areas outside the subset
        return summarize_by_class(np.multiply(self.data, selector),
            np.multiply(class_array, selector), **kwargs)

    def timestamp(self):
        '''
        Returns the ISO 8601 timestamp of the file.

        Returns
        -------
        datetime.date
        '''
        cls = self.__class__
        match = cls.date_regex.match(os.path.dirname(self.filename))
        if match is None:
            return match

        return cls.ordinal_to_date(*map(int, match.groups()))


class TCFArray(SparseArray, metaclass = MetaSparseArray):
    '''
    Represents a "sparse" or "land" array, output from the TCF C codebase.

    Parameters
    ----------
    file_path : str
        File system path to a stored output TCF array in "sparse" form;
        assumes that the filename incorporates the data type, e.g.,
        `foo.flt32` or `foo.uint8`.
    '''
    file_regex = re.compile(r'^tcf_.*_(?P<field>Ctot|gpp|nee|rhtot)_avg_(?P<pft>pft\d{1})?_?.*\.(?P<ftype>%s$)' % '|'.join(TYPE_MAP.keys()))
    date_regex = re.compile(r'.*/Y(?P<year>\d{4})/D(?P<day>\d{3})/?$')

    def __init__(self, file_path, grid = 'M09'):
        super(TCFArray, self).__init__(file_path, grid)
        match = self.file_regex.match(os.path.basename(self.filename))
        self._field, self._pft, self._ftype = match.groups()
        self._field_l4c = FIELD_MAP[self._field] # L4C field name



'''
Essential functions and classes for calibration. See, in particular:

- `OPT_BOUNDS`, which has lower and upper bounds on calibration parameters

**You must create a configuration JSON file before calibrating L4C.** There
is a template available in the directory:

    pyl4c/data/fixtures/files

The optimization routines here are best used with multiple `--trials`, i.e.,
multiple, random initial parameter values. This can help avoiding falling into
a local optimum when a better solution set is available. However, if you are
repeatedly seeing the message:

    Error in objective function; restarting...

Most likely, the bounds on your parameters are producing initial parameter
values that are totally unreasonable.

Note that calibration of L4C (i.e., updating the BPLUT) can be performed using
the command-line interface in `pyl4c.apps.calibration.main`, for example:

    # Build the scratch dataset needed for calibration
    python main.py setup

    # You can get a preview of what filtering the data would look like...
    python main.py pft <pft_number> filter-preview gpp <window_size>
    python main.py pft <pft_number> filter-preview reco <window_size>

    # Optionally, filter the tower datasets to remove spurious spikes
    python main.py filter-all gpp <window_size>
    python main.py filter-all reco <window_size>

    # Run the GPP calibration for a given Plant Functional Type (PFT)
    python main.py pft <pft_number> tune-gpp

    # Run the RECO calibration for a given Plant Functional Type (PFT)
    python main.py pft <pft_number> tune-reco

    # Finally, to dump the updated BPLUT into a CSV or Python pickle file...
    python main.py bplut pickle <output_path> --version-id=<version_id>

    # Get help on any command with --help, for example:
    python main.py plot-gpp --help
'''

import os
import pickle
import nlopt
import netCDF4 # Necessary to import this before HDF5 due to a bug
import h5py
import numpy as np
from collections import OrderedDict
from scipy import optimize
from pyl4c import suppress_warnings
from pyl4c.stats import detrend, rmsd, sum_of_squares
from pyl4c.science import k_mult

# Constrained optimization bounds
OPT_BOUNDS = {
    'gpp': ( # lue, tmin0, tmin1, vpd0, vpd1, smrz0, smrz1, ft0
        np.array((0.5, 230, 276,    0,  1501,  0,  30.1, 0.)), # Lower bound
        np.array((4.0, 275, 320, 1500, 10000, 30, 100,   1.))), # Upper bound
    'reco': ( # CUE, beta_tsoil, smsf0, smsf1
        np.array((0.0,   1,    0,  25)),
        np.array((0.7, 800, 24.9, 100)))
}


class BPLUT(object):
    '''
    Represents a Biome Properties Look-Up Table (BPLUT) with PFT classes along
    the rows and parameters along the columns.

    If initialized with a `params_dict`, these are the values of the BPLUT.
    If initialized with an `hdf5_path` but without a `params_dict`, the
    parameters are read-in from the HDF5 file.
    '''
    _labels = [
        'LUE', 'CUE', 'tmin0', 'tmin1', 'vpd0', 'vpd1', 'smrz0', 'smrz1',
        'smsf0', 'smsf1', 'ft0', 'ft1', 'tsoil', 'decay_rates0',
        'decay_rates1', 'decay_rates2', 'f_metabolic', 'f_structural'
    ]
    _npft = 10 # Number of (possible) PFT classes
    _valid_pft = range(1, 9) # Canonical range of valid PFTs

    def __init__(
            self, params_dict = None, labels = None, hdf5_path = None,
            hdf5_group = 'BPLUT'):
        '''
        Parameters
        ----------
        params_dict : dict
            A BPLUT to initialize the new BPLUT
        labels : tuple or list
            Names of the parameters
        hdf5_path : str
            Path to an HDF5 file to use as a temporary store
        hdf5_group : str
            Field name with which to store data in HDF5 file
            (Default: "BPLUT")
        '''
        self.hdf5_group = hdf5_group
        self.hdf5_path = hdf5_path
        if labels is not None:
            print('WARNING: Parameter names ending with a number are assumed to be bounds on ramp functions!')
            self._labels = labels
        # Create an in-memory parameter dictionary
        empty = self._empty_dict(self.labels)
        if params_dict is None:
            init_data = empty # No prior BPLUT, use empty table
        else:
            init_data = params_dict.copy()
            # IMPORTANT: Make sure prior BPLUT has all the necessary params
            for key in empty.keys():
                init_data.setdefault(key, empty[key])
        # Optionally, maintain a file BPLUT
        if hdf5_path is not None:
            # Restore from the file data, filling in NaNs with initial
            if os.path.exists(hdf5_path):
                with h5py.File(hdf5_path, 'r') as hdf:
                    if hdf5_group in hdf.keys():
                        init_data = self.hdf5_restore(hdf, init_data = init_data)
            # Then, if file dataset doesn't exist, create a new one and store
            #   the initial data
            with h5py.File(hdf5_path, 'a') as hdf:
                self.hdf5_flush(hdf, data = init_data)
        self.data = init_data

    @property
    def data(self):
        'The parameters dictionary or `dict` instance'
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def labels(self):
        'Names of the free parameters'
        return self._labels

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def _empty_dict(self, labels = None, dtype = np.float32):
        '''
        Given sequence of labels, convert to legacy, human-readable dict,
        e.g.:

            { 'LUE': array([[ nan, 1.17, ..., nan ]]), ... }
        '''
        labels_dedupe = self._canonical(labels)
        result = dict()
        for name in labels_dedupe:
            size = len(list(filter(lambda x: x.startswith(name), labels)))
            result[name] = np.ones((size, self._npft), dtype = dtype) * np.nan
        return result

    def _canonical(self, labels = None):
        '''
        Returns a short list of labels, without duplicates. Specifically,
        a list of labels like, e.g., ("tmin0", "tmin1", "decay_rates1")
        becomes ("tmin", "decay_rates").

        Parameters
        ----------
        labels : tuple or list or None

        Returns
        -------
        list
        '''
        labels = self.labels if labels is None else labels
        # Remove numeric suffixes and de-duplicate the list
        return list(OrderedDict([
            (p.strip('0123456789'), 0) for p in labels
        ]).keys())

    def flat(self, pft, labels = None):
        '''
        Retrieves a flat list of parameters for a specific PFT.

        Parameters
        ----------
        pft : int
            Numeric code of the PFT for which to return parameter values
        labels : tuple or list
            (Optional) A sequence of parameter names desired, if not all;
            defaults to returning all parameters

        Returns
        -------
        numpy.ndarray
        '''
        labels_dedupe = self._canonical(labels)
        return np.hstack([
            self.data[p][:,pft].ravel() if p in self.data.keys() else np.nan
            for p in labels_dedupe
        ])

    def hdf5_flush(self, hdf, data = None):
        '''
        Writes the current BPLUT to an HDF5 file.

        Parameters
        ----------
        hdf : h5py.File
            HDF5 file open for writing
        data : dict
        '''
        assert hdf.mode != 'r', 'File not open for writing!'
        data = self.data if data is None else data
        if self.hdf5_group not in hdf.keys():
            hdf.create_group(self.hdf5_group)
        for key, value in data.items():
            if key.startswith('_'):
                continue # Skip keys that are not parameter names
            field = '%s/%s' % (self.hdf5_group, key)
            if key not in hdf[self.hdf5_group].keys():
                hdf.create_dataset(field, value.shape, np.float32, value)
            else:
                # Overwrite NaNs in the file data
                _value = hdf[field][:]
                hdf[field][:] = np.where(np.isnan(_value), value, _value)
        hdf.flush()

    def hdf5_restore(self, hdf, init_data = None, dtype = np.float32):
        '''
        Reads in the BPLUT table stored in the HDF5 file.

        Parameters
        ----------
        hdf : h5py.File
        init_data : dict
            Initital data; will be over-written by HDF5 file contents

        Returns
        -------
        dict
        '''
        data = dict() if init_data is None else init_data
        for key in hdf[self.hdf5_group].keys():
            # Update the in-memory BPLUT
            from_hdf5 = hdf[self.hdf5_group][key][:]
            data[key] = np.where(
                ~np.isnan(from_hdf5), from_hdf5, init_data.get(key)
            ).astype(dtype)
        return data

    def pickle(self, output_path, version_id = None):
        '''
        Writes the current BPLUT parameters, as a dictionary, to a pickle
        file.

        Parameters
        ----------
        output_path : str
            The output path for the pickle file (*.pickle)
        version_id : str
            (Optional) The version identifier for this BPLUT
        '''
        with open(output_path, 'wb') as file:
            output = self.data.copy()
            if version_id is not None:
                output['_version'] = version_id
            pickle.dump(output, file)

    def show(self, pft, param, precision = 2):
        '''
        Prints the current BPLUT parameters for a given PFT, the values of a
        given parameter for all PFTs, or the value of a specific PFT-parameter
        combination.

        Parameters
        ----------
        pft : int or None
            The PFT class
        param : str or None
            The name of the parameter
        precision : int
            Decimal precision to use for printing numbers
        '''
        assert not (pft is None and param is None),\
            'Either one or both must be specified: --pft or --param'
        set_of_labels = self._canonical() if param is None else [param]
        set_of_pfts = self._valid_pft if pft is None else [pft]
        for each in set_of_labels:
            assert each in self._canonical(), 'Unrecognized parameter: %s' % each
        for each in set_of_pfts:
            assert each in range(0, self._npft), 'PFT code out of range'
        pad = max(len(l) for l in set_of_labels) + 2
        fmt_string = '{:>%d} {:>%d}' % (pad, 5 + precision)
        for pft in set_of_pfts:
            print('BPLUT parameters for PFT %d:' % pft)
            for label in set_of_labels:
                param_values = self.data[label][:,pft]
                for i, value in enumerate(param_values):
                    # If there are multiple values for a parameter (group),
                    #   append a number to the end of the label
                    if len(param_values) > 1:
                        prefix = '%s%d:' % (label, i)
                    else:
                        prefix = '%s:' % label
                    print(
                        fmt_string.format(prefix, ('%%.%df' % precision) % value))

    def update(self, pft, values, labels, flush = True):
        '''
        Updates the BPLUT with the specified parameters for a single PFT.

        Parameters
        ----------
        pft : int
            The PFT class
        values : tuple or list
            Sequence of parameter values, one for each parameter named in
            `labels`
        labels : tuple or list
            Sequence of parameter names, one for each value in `values`
        flush : bool
            True to write the result to disk (attached HDF5 file storage)
            (Default: True)
        '''
        assert len(values) == len(labels),\
            'Vectors of values and parameter labels must have the same length'
        if flush:
            assert self.hdf5_path is not None,\
                'No HDF5 file storage is attached'
            hdf = h5py.File(self.hdf5_path, 'a')
        for i, name in enumerate(labels):
            abbrv = name.strip('0123456789')
            # In case parameter has multiple levels, like a ramp function
            #   (e.g., "smsf0" and "smsf1" are two rows)
            dupes = list(filter(lambda x: x.startswith(abbrv), self.labels))
            j = dupes.index(name) # If it has one level (e.g., "LUE"), j = 0
            self.data[abbrv][j,pft] = values[i]
            if flush:
                path = '%s/%s' % (self.hdf5_group, abbrv)
                hdf[path][:,pft] = self.data[abbrv][:,pft]
        if flush:
            hdf.flush()
            hdf.close()


class ModelParameters(OrderedDict):
    '''
    Convenience wrapper for an OrderedDict, allowing both vectorized and
    keyword access to model parameters.

    Parameters
    ----------
    group : str
        Name of this model parameters group, usually the name of the model or
        sub-model to which they belong
    *params : spotpy.parameter
        One or more parameters
    '''
    def __init__(self, group, *params):
        self._group = group
        # Create {name: spotpy.parameter, ...} dictionary
        super().__init__(**dict([(p.name, p) for p in params]))


class GenericOptimization(object):
    '''
    A more generic and expansive tool for optimization; includes many more
    algorithms for minimization/ maximization problems, including sequential
    quadratic programming (SQP), which is the default here and is closest to
    what is performed in Matlab's `fmincon`. Despite the similarity to `fmincon`,
    SQP will tend to deviate strongly from the initial parameters derived via
    fmincon. This solver is SLOW for gradient descent methods relative to
    `scipy.optimize.least_squares()`, because the gradient is calculated with
    a finite element approach.

        opt = GenericOptimization(residuals, OPT_BOUNDS['gpp'],
            step_size = (0.01, 0.1, 0.1, 1, 1, 0.1, 0.1, 0.05))
        opt.solve(init_params)

    See: https://nlopt.readthedocs.io/en/latest/NLopt_Python_Reference/

    Parameters
    ----------
    func : function
        Function to calculate the residuals
    bounds : list or tuple
        2-element sequence of (lower, upper) bounds where each element is an
        array
    method : str
        One of the nlopt algorithms
    step_size : list or tuple or numpy.ndarray
        Sequence of steps to take in gradient descent; not needed for
        derivative-free methods
    verbose : bool
        True to print all output to the screen
    '''
    def __init__(
            self, func, bounds, method: int = nlopt.LD_SLSQP,
            step_size = None, verbose = True):
        # https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/#slsqp
        assert isinstance(method, int), 'Did not recognize "method" argument'
        self._bounds = bounds
        self._method = method
        self._residuals = func
        self._step_size = step_size
        self._verbose = verbose

    def solve(self, init_params, ftol = 1e-8, xtol = 1e-8, maxeval = 500):
        '''
        Using the sum-of-squared errors (SSE) as the objective function,
        solves a minimization problem.

        Parameters
        ----------
        init_params : list or tuple or numpy.ndarray
            Sequence of starting parameters (or "initial guesses")
        ftol : float
        xtol : float
        maxeval : int
            Maximum number of objective function evaluations

        Returns
        -------
        numpy.ndarray
        '''
        @suppress_warnings
        def sse(x):
            return np.power(self._residuals(x), 2).sum()

        @suppress_warnings
        def objf(x, grad):
            if grad.size > 0:
                # Approximate the gradient using finite element method
                grad[...] = optimize.approx_fprime(
                    x, sse, self._step_size)
            return sse(x)

        opt = nlopt.opt(self._method, len(init_params))
        # https://nlopt.readthedocs.io/en/latest/NLopt_Python_Reference/#localsubsidiary-optimization-algorithm
        if self._method == nlopt.G_MLSL_LDS:
            opt.set_local_optimizer(
                nlopt.opt(nlopt.LN_COBYLA, len(init_params)))
        opt.set_min_objective(objf)
        opt.set_lower_bounds(self._bounds[0])
        opt.set_upper_bounds(self._bounds[1])
        opt.set_ftol_abs(ftol)
        opt.set_xtol_abs(xtol)
        opt.set_maxeval(maxeval)
        if self._verbose:
            print('Solving...')
        return opt.optimize(init_params)


def cbar(rh, k_mult, q_rh = 75, q_k = 50):
    '''
    Calculates "Cbar," the time-constant upper quantile of the RH/Kmult
    ratio. Where Kmult is >/= `q_k`, return the `q_rh` quantile of RH/Kmult;
    intended for T x N arrays where T is the number of time steps and
    N is the number of (flux tower) sites.

    Parameters
    ----------
    rh : numpy.ndarray
        (T x N) vector of heterotrophic respiration
    k_mult : numpy.ndarray
        (T x N) vector of Kmult
    q_rh : float
        Percentile of RH/Kmult to return
    q_k : float
        Percentile of Kmult below which RH/Kmult values are masked

    Returns
    -------
    numpy.float64
    '''
    cutoff = np.apply_along_axis(
        np.percentile, 0, k_mult, q = q_k).reshape((1, k_mult.shape[1]))
    return np.nanpercentile(
        np.where(k_mult >= cutoff,
            np.divide(rh, np.where(k_mult == 0, np.nan, k_mult)), np.nan),
        q = q_rh, axis = 0)


def reco(params, tsoil, smsf, reco_tower, gpp_tower, q_rh = 75, q_k = 50):
    '''
    Calculate empirical ecosystem respiration, RECO, based on current model
    parameters and the inferred soil organic carbon (SOC) storage; i.e., this
    calculation should be used in model calibration when SOC is not a priori
    known, see `pyl4c.apps.calibration.cbar()`. The expected model parameter
    names are "CUE" for the carbon use efficiency of plants.

    Parameters
    ----------
    params : dict
        A dict-like data structure with named model parameters
    tsoil : numpy.ndarray
        (T x N) vector of soil temperature (deg K), where T is the number of
        time steps, N the number of sites
    smsf : numpy.ndarray
        (T x N) vector of surface soil wetness (%), where T is the number of
        time steps, N the number of sites
    reco_tower : numpy.ndarray
        (T x N) vector of observed RECO from eddy covariance tower sites
    gpp_tower : numpy.ndarray
        (T x N) vector of observed GPP from eddy covariance tower sites
    q_rh : int
        The percentile of RH/Kmult to use in calculating Cbar
    q_k : int
        The percentile of Kmult below which RH/Kmult values are masked

    Returns
    -------
    numpy.ndarray
    '''
    # Calculate RH as (RECO - RA) or (RECO - (faut * GPP));
    #   globals "reco_tower", "gpp_tower"
    ra = ((1 - params['CUE']) * gpp_tower)
    rh = reco_tower - ra
    rh = np.where(rh < 0, 0, rh) # Mask out negative RH values
    # Compute Cbar with globals "q_rh" and "q_k"
    kmult0 = k_mult(params, tsoil, smsf)
    cbar0 = cbar(rh, kmult0, q_rh, q_k)
    return ra + (kmult0 * cbar0)


def report_fit_stats(obs, pred, weights = np.array([1]), verbose = True):
    '''
    Reports the RMSE, ubRMSE, and Bias for observed and predicted values.

    Parameters
    ----------
    obs : numpy.ndarray
        Vector of observed ("true") values
    pred : numpy.ndarray
        Vector of predicted values
    weights : numpy.ndarray
        (Optional) Vector of weights for each sample

    Returns
    -------
    tuple
        (R-squared, RMSE, ubRMSE, Bias)
    '''
    y = np.apply_along_axis(detrend, 0, obs, fill = True)
    yhat = np.apply_along_axis(detrend, 0, pred, fill = True)
    rmse = rmsd(obs, pred, weights = weights)
    ubrmse = rmsd(y, yhat, weights = weights)
    bias = np.nanmean(np.subtract(pred, obs))
    mask = np.logical_or(np.isnan(obs), np.isnan(pred))
    r_squared = 1 - np.divide(
        sum_of_squares(
            obs[~mask], pred[~mask], add_intercept = False, which = 'sse'),
        sum_of_squares(
            obs[~mask], pred[~mask], add_intercept = False, which = 'sst'))
    if verbose:
        print('Fit statistics:')
        print('--    R^2: %s' % ('%.3f' % r_squared).rjust(6))
        print('--   RMSE: %s' % ('%.3f' % rmse).rjust(6))
        print('-- ubRMSE: %s' % ('%.3f' % ubrmse).rjust(6))
        print('--   Bias: %s' % ('%.3f' % bias).rjust(6))
    return (r_squared, rmse, ubrmse, bias)


def solve_least_squares(func, init_params, labels, bounds, **kwargs):
    '''
    Apply constrained, non-linear least-squares optimization. Mostly a
    wrapper for `scipy.optimize.least_squares()`.

    Parameters
    ----------
    func : function
        Function to calculate the residuals
    init_params : list or tuple or numpy.ndarray
        Sequence of starting parameters (or "initial guesses")
    labels : list or tuple or numpy.ndarray
        Sequence of parameter names
    bounds : list or tuple
        2-element sequence of (lower, upper) bounds where each element is an
        array

    Returns
    -------
    scipy.optimize.OptimizeResult
    '''
    # Update the optimization settings; this loss function produces
    #   an estimate for the FT multiplier that is closest to prior
    kwargs.setdefault('loss', 'arctan')
    kwargs.setdefault('method', 'trf')
    kwargs.setdefault('max_nfev', 500)
    kwargs.setdefault('ftol', 1e-8)
    kwargs.setdefault('xtol', 1e-8)
    kwargs.setdefault('gtol', 1e-8)
    try:
        solution = optimize.least_squares(
            func, init_params, bounds = bounds, **kwargs)
    except ValueError:
        below = [
            labels[i]
            for i in np.argwhere(init_params < bounds[0]).flatten().tolist()
        ]
        above = [
            labels[i]
            for i in np.argwhere(init_params > bounds[1]).flatten().tolist()
        ]
        if np.isnan(init_params).any():
            raise ValueError(
                'Error in candidate parameter values; residual function probably returning NaNs')
        else:
            raise ValueError(
                '"Infeasibility" error; check lower bound on %s; upper bound on %s' % (
                '(None)' if len(below) == 0 else ', '.join(below),
                '(None)' if len(above) == 0 else ', '.join(above)))
    return solution



'''
Utilities for working with NetCDF files.
'''

import calendar
import datetime
import re
import numpy as np
import netCDF4
from osgeo import osr
from osgeo import gdal
from osgeo import gdalconst
from bisect import bisect_left
from scipy.io import netcdf_file
from pyl4c.epsg import EPSG
from pyl4c.spatial import EASE2_GRID_PARAMS, array_to_raster, dump_raster
from pyl4c.utils import get_slice_idx_by_bbox

TIME_RX = re.compile(r'^(?P<interval>seconds|hours|days|months|years) since (?P<epoch>[\d\-]+(?:\s{1}?(?:\d{2}\:\d{2}))?(?:\:\d{2})?)')
NDAYS_RX = re.compile(r'(\d+)\_day')

def nc_dim(variable):
    '''
    Returns the dimensions of a netCDF variable.

    Parameters
    ----------
    variable : scipy.io.netcdf.netcdf_variable

    Returns
    -------
    int
    '''
    if hasattr(variable, 'ndim'):
        return variable.ndim
    if hasattr(variable, 'dimensions'):
        return len(variable.dimensions)


def netcdf_array(nc, keys, cell_size = None, time_idx = 0, x_offset = -180):
    '''
    Extracts an equirectangular NetCDF data array as a NumPy array, with
    spatial reference system (SRS) information. NOTE: Both `cell_size` elements
    should be positive numbers.

    Parameters
    ----------
    nc : netCDF4.Dataset
        The NetCDF file (opened)
    keys : tuple
        A tuple of (variable, x_coords, y_coords) where variable is the
        variable name, x_coords the name of the X-coordinate array variable,
        y_coords the name of the Y-coordinate array variable
    cell_size : tuple or None
        (Optional) Tuple of the the cell size in both directions in degrees,
        e.g., (1, 1); will attempt to infer the cell size if not provided
    time_idx : int
        (Optional) Index of the "time" axis to extract as a raster
        (Default: 0)
    x_offset : int or float
        (Optional) Because of the absurd X-coordinates convention used in some
        NetCDF arrays, it may be necessary to "offset" the X coordinates; this
        is only used if it is detected that the maximum X coordinate value is
        >/= 180 degrees E (Default: -180);

    Returns
    -------
    tuple
        Tuple of `(array, gt, wkt)`
    '''
    def infer_cell_size(x_coords, y_coords):
        assert x_coords.ndim == 1,\
            "Can't infer cell size from coordinate arrays with more than one axis"
        # NOTE: Rounding coordinate arrays to 5 decimal places for comparison
        x_diff = np.round(np.array(x_coords)[1:] - np.array(x_coords)[0:-1], 4)
        y_diff = np.round(np.array(y_coords)[1:] - np.array(y_coords)[0:-1], 4)
        assert np.median(x_diff) == x_diff[0] and np.median(y_diff) == y_diff[0],\
            "Array cell size is not equal in both directions!"
        return (np.abs(np.median(x_diff)), np.abs(np.median(y_diff)))

    # Original (equirectangular) spatial reference system
    wkt0 = osr.SpatialReference()
    wkt0.ImportFromEPSG(4326)
    if cell_size is not None:
        x_res, y_res = cell_size
        assert x_res > 0 and y_res > 0, 'Argument cell_size must be greater than zero'
    else:
        x_res, y_res = infer_cell_size(
            np.array(nc.variables[keys[1]][:]),
            np.array(nc.variables[keys[2]][:]))

    x0 = np.array(nc.variables[keys[1]][:]).min()
    y0 = np.array(nc.variables[keys[2]][:]).max()
    # Geotransform params: (x_min, pixel_width, 0, y_max, 0, -pixel_height)
    gt0 = (x0, x_res, 0, y0, 0, -y_res)

    # NOTE: Because some NetCDF files have a "degrees east" convention with
    #   longitude increasing monotonically from 0 to 360, we must apply an
    #   offset to the X coordinates
    start_idx = None
    if np.array(nc.variables[keys[1]][:]).max() > 180:
        # Update affine transformation
        gt0 = (x0 + x_offset, x_res, 0, y0, 0, -y_res)
        # Transform to proper longitudes
        x_coords = (nc.variables[keys[1]][:] + x_offset)

        # Get X-coordinate origin, or index of east longitude closest to zero,
        #   unless all coordinates are in a single hemisphere
        test = np.logical_and(
            x_coords > 0, x_coords <= np.abs(x_coords).min())
        if np.any(test):
            idx = np.arange(0, x_coords.shape[0])
            start_idx = int(idx[x_coords == x_coords[test]])

    # Extract a NumPy array
    if hasattr(time_idx, 'index'):
        a, b = (min(time_idx), max(time_idx))
        arr = np.array(nc.variables[keys[0]][a:(b+1),...])
    else:
        # Make sure a single time slice as a band axis
        if nc_dim(nc.variables[keys[0]]) > 2:
            arr = np.array(nc.variables[keys[0]][time_idx,...])[np.newaxis,...]
        else:
            # Some netCDF variables may have no time/ band axis (i.e., static)
            arr = np.array(nc.variables[keys[0]][:])[np.newaxis,...]

    # If Y coordinates are in ascending order, flip the array
    lats = nc.variables[keys[2]][:].tolist()
    if lats.index(max(lats)) > lats.index(min(lats)):
        arr = np.flip(arr, axis = 1)

    # If necessary, re-sort data so west-most longitude is the first column
    if start_idx is not None:
        arr = np.concatenate((arr[...,start_idx:], arr[...,:start_idx]), axis = 2)

    scale = 1
    if hasattr(nc.variables[keys[0]], 'scale_factor'):
        scale = nc.variables[keys[0]].scale_factor

    offset = 0
    if hasattr(nc.variables[keys[0]], 'add_offset'):
        offset = nc.variables[keys[0]].add_offset

    return ((arr * scale) + offset, gt0, wkt0.ExportToWkt())


def netcdf_raster(
        nc, keys, cell_size = None, time_idx = 0, subset_bbox = None,
        in_nodata = None, out_nodata = -9999):
    '''
    Dumps a NetCDF variable to a GeoTIFF file. NOTE: Both `cell_size` elements
    should be positive numbers.

    Parameters
    ----------
    nc : netCDF4.Dataset
        The NetCDF file (opened)
    keys : tuple
        A tuple of (variable, x_coords, y_coords) where variable is the
        variable name, x_coords the name of the X-coordinate array variable,
        y_coords the name of the Y-coordinate array variable
    cell_size : tuple or None
        (Optional) Tuple of the the cell size in both directions in degrees,
        e.g., (1, 1); will attempt to infer the cell size if not provided
    time_idx : int
        (Optional) Index of the "time" axis to extract as a raster
        (Default: 0)
    subset_bbox : tuple or None
        (Optional) Subset bounding box
    in_nodata : int or float or None
        (Optional) The NoData value to ignore in the input
    out_nodata : int or float or None
        (Optional) The NoData value to set in the output (Default: -9999)

    Returns
    -------
    gdal.Dataset
    '''
    arr, gt0, wkt0 = netcdf_array(nc, keys, cell_size, time_idx)
    xmin = ymin = 0
    if subset_bbox is not None:
        x_coords = nc.variables[keys[1]][:]
        y_coords = nc.variables[keys[2]][:]
        if np.array(nc.variables[keys[1]][:]).max() > 180:
            # Correct for weird NetCDF easting
            x_coords = nc.variables[keys[1]][:] - 180
        x_idx, y_idx = get_slice_idx_by_bbox(
            x_coords, y_coords, subset_bbox = subset_bbox)
        xmin, xmax = x_idx
        ymin, ymax = y_idx
        if arr.ndim == 2:
            arr = arr[ymin:ymax, xmin:xmax]
        else:
            arr = arr[:, ymin:ymax, xmin:xmax]

    # Optional: Mask user-specified NoData value
    if in_nodata is not None:
        arr = np.where(arr == in_nodata, np.nan, arr)

    # Mask any NoData/ fill value defined by the dataset
    if hasattr(nc.variables[keys[0]], 'missing_value'):
        arr = np.where(
            arr == nc.variables[keys[0]].missing_value, np.nan, arr)

    # Create a gdal.Dataset from the array
    if out_nodata is not None:
        return array_to_raster(
            np.where(
                np.isnan(arr), out_nodata, arr), gt0, str(wkt0), xmin, ymin)

    return array_to_raster(arr, gt0, str(wkt0), xmin, ymin)


def parse_time_units(netcdf_time_axis):
    '''
    Returns the time interval size (e.g., "days") and the epoch
    (e.g., "1970-01-01") from a given netCDF units declaration for the time
    axis (e.g., "days since 1970-01-01").

    Parameters
    ----------
    netcdf_time_axis : netCDF4._netCDF4.Variable
        Variable that describes the time axis

    Returns
    -------
    tuple
        Tuple of (str, datetime.datetime)
    '''
    units_string = netcdf_time_axis.units
    if isinstance(units_string, bytes):
        units_string = bytes(units_string).decode()

    interval, epoch_str = TIME_RX.match(units_string).groups()
    epoch_str = epoch_str.replace('-1-1', '-01-01')
    if epoch_str.rfind('00:00:00') > 0:
        epoch = datetime.datetime.strptime(epoch_str, '%Y-%m-%d %H:%M:%S')
    elif epoch_str.rfind('00:00') > 0:
        epoch = datetime.datetime.strptime(epoch_str, '%Y-%m-%d %H:%M')
    else:
        epoch = datetime.datetime.strptime(epoch_str, '%Y-%m-%d')
    return (interval, epoch)


def spatial_average(
        nc, keys, reducer, t_labels = None, subset_bbox = None, nodata = -9999):
    '''
    Calculates a spatial average across a NetCDF time series variable.

    Parameters
    ----------
    nc : netCDF4.Dataset
        The NetCDF file (opened)
    keys : tuple
        A tuple of (variable, x_coords, y_coords) where variable is the
        variable name, x_coords the name of the X-coordinate array variable,
        y_coords the name of the Y-coordinate array variable
    reducer : function
        A vectorized function that returns a scalar for an input vector of
        values; should ignore np.nan values
    t_labels : tuple or list
        (Optional) A vector of time index labels
    subset_bbox : tuple
        (Optional) To specify a subset area for the spatial average (i.e.,
        instead of averaging over the entire NetCDF array domain), pass a
        bounding box argument
    nodata : int or float
        The NoData value to ignore (Default: -9999)

    Returns
    -------
    list
        Of the form `[(index, value), ...]` for each time step
    '''
    if subset_bbox is not None:
        x_coords = nc.variables[keys[1]][:]
        y_coords = nc.variables[keys[2]][:]
        if np.array(nc.variables[keys[1]][:]).max() > 180:
            # Correct for weird NetCDF easting
            x_coords = nc.variables[keys[1]][:] - 180
        x_idx, y_idx = get_slice_idx_by_bbox(
            x_coords, y_coords, subset_bbox = subset_bbox)
        xmin, xmax = x_idx
        ymin, ymax = y_idx

    scale = 1
    if hasattr(nc.variables[keys[0]], 'scale_factor'):
        scale = nc.variables[keys[0]].scale_factor

    offset = 0
    if hasattr(nc.variables[keys[0]], 'add_offset'):
        offset = nc.variables[keys[0]].add_offset

    num_epochs = nc.variables[keys[0]].shape[0]
    time_series = []
    for i, t in enumerate(range(0, num_epochs)):
        if subset_bbox is not None:
            arr = nc.variables[keys[0]][:][t, ymin:ymax, xmin:xmax]

        else:
            arr = nc.variables[keys[0]][:][t,...]

        index = t
        if t_labels is not None:
            index = t_labels[i]

        time_series.append((
            # Filter out NoData; use summary function that ignores NaNs
            index, reducer( # Multiply by scale, add offset
                np.where(arr == nodata, np.nan, (arr * scale) + offset))
        ))

    return time_series


def time_series(netcdf_time_axis, has_leap = None):
    '''
    Constructs a series of datetime.datetime elements based on a netCDF
    time axis; assumes that time steps are evenly spaced (except in leap
    years) and ignores fractional time steps.

    NOTE: This is EXTREMELY difficult to get right because of awful
    implementation of time axes in the typical netCDF dataset. Among other
    challenges... Apparently, netCDF time axes use the convention that if the
    epoch starts at midnight, the first day is the "zeroth" day; e.g., CLM5.0
    states its time axis is "days since 1700-01-01 00:00:00" but the first
    time step is 31.0; it seems highly unlike this monthly time series would
    start in February (Jan. 1 + 31 days == Feb. 1), so they must be using the
    *last day* of each month (e.g., Jan 0 + 31 days == Jan. 31).

    Parameters
    ----------
    netcdf_time_axis : netCDF4._netCDF4.Variable
        Variable that describes the time axis
    has_leap : bool or None
        Set True to force recognition of leap years; if None, will attempt
        to automatically determine whether leap years should be recognized
        based on the field metadata (Default: None)

    Returns
    -------
    list
        `[*datetime.date|datetime.datetime]`
    '''
    steps = netcdf_time_axis[:]
    interval, epoch = parse_time_units(netcdf_time_axis)
    convention = getattr(netcdf_time_axis, 'calendar', '')
    if hasattr(convention, 'decode'):
        convention = convention.decode('utf-8')

    # Attempt to determine leap year recognition
    if has_leap is None:
        has_leap = convention in ('', 'standard', 'gregorian')

    first_step = steps[1] - steps[0]
    # Basically, we can tolerate different step sizes for intervals that
    #   can be passed to datetime.timedelta, which assumes leap years;
    #   in all other cases, we can't calculate time steps correctly
    if not has_leap:
        if not np.all(np.array(steps[1:]) - np.array(steps[:-1]) == first_step):
            assert interval in ('seconds', 'hours', 'days'),\
                'Time series is not evenly spaced and leap years do not explain this discrepancy'

    ldays = 1 if has_leap else 0
    ndays = 365
    if NDAYS_RX.match(convention) is not None:
        # Some datasets oddly have, e.g., "360_day" conventions
        ndays = int(NDAYS_RX.match(convention).groups()[0])

    # Because timedelta can't process intervals higher than "days"...
    if interval == 'years':
        # Simple years series starting January 1 if units are years
        years = np.arange(epoch.year, epoch.year + steps[-1] + 1)
        time_axis = [datetime.date(int(y), 1, 1) for y in years]
    elif interval == 'months':
        time_axis = []
        m_range = (np.arange(0, len(steps)) + epoch.month) % 12
        m_range = np.where(m_range == 0, 12, m_range)
        for t, t_value in enumerate(steps):
            m = m_range[t]
            y = epoch.year + np.floor(t_value / 12)
            time_axis.append(datetime.date(int(y), int(m), epoch.day))
    # But if time intervals are in "seconds," "hours," or "days"
    elif interval in ('seconds', 'hours', 'days') and has_leap:
        time_axis = [
            epoch +
                datetime.timedelta(**{interval: int(t)}) for t in steps
        ]
    # But timedelta assumes leap years are in use, so if the time series
    #   doesn't use leap years, we must construct it manually
    else:
        time_axis = []
        this_year = epoch.year
        mdays = np.cumsum(calendar.mdays)
        for t in steps:
            if interval == 'seconds':
                y = epoch.year + (t / (ndays * 24 * 60 * 60))
            if interval == 'hours':
                y = epoch.year + (t / (ndays * 24))
            if interval in ('seconds', 'hours'):
                m = bisect_left(mdays, ((y - epoch.year) * ndays) % ndays)
                d = (((y - epoch.year) * ndays) % ndays) - mdays[m - 1]
            if interval == 'days':
                y = epoch.year + (t / ndays) # e.g., t / 365
                # Incidentally, the left "sorting index" is the month number
                m = bisect_left(mdays, t % ndays)
                m += 1 if m == 0 else 0
                d = (t % ndays) - mdays[m - 1] # And days are what is left over
                d += 1 if d == 0 else 0
            time_axis.append(datetime.datetime(int(y), int(m), int(d)))
    return time_axis



'''
Classes for building command line interface (CLI) tools, such as for bulk
processing data at the command line.
'''

import os
import sys
import h5py
import numpy as np
from osgeo import gdal
from cached_property import cached_property
from pyl4c.data.fixtures import EASE2_GRID_PARAMS
from pyl4c.epsg import EPSG
from pyl4c.spatial import array_to_raster, as_array, ease2_coords_approx
from pyl4c.utils import get_ease2_coords, get_ease2_slice_idx
from pyl4c.lib.tcf import TYPE_MAP, SparseArray, TCFArray

class CommandLineInterface(object):
    '''
    A command line interface (CLI) convenience class used for creating
    Python scripts that can be invoked from the command line.
    '''

    def __init__(self):
        pass

    def __check__(self):
        possible_required_keys = ('_output_path', '_output_tpl')
        for key in possible_required_keys:
            if hasattr(self, key):
                assert getattr(self, key) is not None, 'You must specify %s with: --%s=""' % (key, key)

        if hasattr(self, '_field_tpl'):
            assert self._field_tpl.rfind('%d') > 0, 'The field argument must contain a string formatting character, e.g., "SOC/soc_pft%d_mean"'

        if hasattr(self, '_grid'):
            assert self._grid is not None, 'Must specify the EASE-Grid 2.0 size with --grid argument ("M01" or "M09")'

        if hasattr(self, '_mask'):
            if self._mask is not None:
                assert isinstance(self._mask, str), 'Did not recognize --mask as a file path'
                assert os.path.exists(self._mask), 'File not found: %s (Cannot use shortcuts like ~)' % self._mask

        if hasattr(self, '_output_dir'):
            assert os.path.isdir(self._output_dir), 'Directory does not exist: %s' % self._output_dir

        if hasattr(self, '_output_path'):
            assert os.path.exists(os.path.dirname(self._output_path)), 'Did not recognize output_path (Cannot use shortcuts like ~)'

        if hasattr(self, '_output_tpl'):
            assert self._output_tpl.rfind('%s') >= 0, 'The output_tpl argument must have one string formatting character'

        if hasattr(self, '_reference'):
            if self._reference is not None:
                assert isinstance(self._reference, str), 'Did not recognize --reference as a file path'
                assert os.path.exists(self._reference), 'File not found: %s (Cannot use shortcuts like ~)' % self._reference

        if hasattr(self, '_summaries'):
            assert not isinstance(self._summaries, str), 'Could not interpret --summaries argument as a sequence of NumPy functions'

    @cached_property
    def __coords__(self):
        if self._grid in ('M01', 'M09'):
            return get_ease2_coords(self._grid)

        # But if coordinate arrays are not pre-computed...
        return ease2_coords_approx(self._grid)

    @cached_property
    def __shp__(self):
        if self._subset_id is not None:
            x_idx, y_idx = self.__slice_idx__
            return (y_idx[1] - y_idx[0], x_idx[1] - x_idx[0])

        x_coords, y_coords = self.__coords__
        return (len(y_coords), len(x_coords))

    @cached_property
    def __slice_idx__(self):
        # The ((xmin, xmax), (ymin, ymax)) indices in pixel space
        if self._subset_id is not None:
            return get_ease2_slice_idx(
                grid = self._grid, subset_id = self._subset_id)

        return (None, None)

    def __pieces__(self):
        # Returns the next row chunk
        p = 0 # Initialize at piece 0
        start = 0 # Starting at first row, if no subsetting
        if self._subset_id is not None:
            x_idx, y_idx = self.__slice_idx__
            start = y_idx[0] # Starting at top of subset

        num_pieces = self._pieces
        num_rows = self.__shp__[0]
        step = int(np.ceil(num_rows / num_pieces))
        while p < num_pieces:
            yield (start + (p * step), start + ((p+1) * step))
            p += 1

    def gdt_to_dtype(self, gdt):
        return {
            gdal.GDT_Float32: np.float32,
            gdal.GDT_Float64: np.float64,
            gdal.GDT_Int16: np.int16,
            gdal.GDT_Int32: np.int32
        }[gdt]

    def lookup_dtype(self, type_string):
        '''
        Given, e.g., "float32", returns `numpy.float32`.

        Parameters
        ----------
        type_string : str
            A NumPy named type, e.g., "float32", "int16", "byte"
        '''
        return getattr(np, type_string)

    def lookup_gdt(self, type_string):
        '''
        Given, e.g., "float32", returns `gdal.GDT_Float32`.

        Parameters
        ----------
        type_string : str
            A NumPy named type, e.g., "float32", "int16", "byte"
        '''
        return getattr(
            gdal, 'GDT_%s' % type_string.title() if type_string != 'uint8' else 'GDT_Byte')

    def infer_file_mode(self, file_paths):
        '''
        Determine what class of file we're working with.

        Parameters
        ----------
        file_path : str

        Returns
        -------
        str
            One of: "hdf5", "sparse", "other"
        '''
        # Check if a list/tuple given versus a character sequence
        path = file_paths[0] if len(file_paths[0]) > 1 else file_paths
        if path.split('.')[-1] == 'h5':
            return 'hdf5'
        elif path.split('.')[-1] in TYPE_MAP.keys():
            return 'sparse'
        else:
            return 'other'

    def read_array(self, file_path, **kwargs):
        '''
        Reads in the file at the given file path and returns a
        `numpy.ndarray`.

        Parameters
        ----------
        file_path : str
            The location of a file
        start_row : int
            Index of the first row in a row chunk
        end_row : int
            Index of the final row in a row chunk
        mode : str
            The file mode to use
        shp : tuple
            The shape of the raster arrays

        Returns
        -------
        numpy.ndarray
        '''
        if 'start_row' in kwargs.keys() and 'end_row' in kwargs.keys():
            if kwargs['start_row'] is not None and kwargs['end_row'] is not None:
                return self.read_chunked(file_path, **kwargs)
        if 'shp' in kwargs.keys():
            shp = kwargs['shp']
        else:
            shp = None

        mode = self._mode
        if 'mode' in kwargs.keys():
            mode = kwargs['mode']
        if self._subset_id is not None:
            x_idx, y_idx = self.__slice_idx__
            xmin, xmax = x_idx
            ymin, ymax = y_idx
        # If we're reading HDF5 files...
        if mode == 'hdf5':
            with h5py.File(file_path, 'r') as hdf:
                assert self._field in hdf.keys(), 'Could not find the specified field name: %s' % self._field
                # If we're not subsetting...
                if self._subset_id is None:
                    return hdf[self._field][:]
                return hdf[self._field][ymin:ymax, xmin:xmax]
        # If we're reading sparse TCF output arrays...
        elif mode in ('sparse', 'tcf'):
            # TCFArray has stronger assumptions than SparseArray; try to read
            #   the file either way
            tcf = SparseArray(file_path, self._grid)
            tcf.inflate()
            if self._subset_id is None:
                return tcf.data[:]
            return tcf.data[ymin:ymax, xmin:xmax]
        # If we're reading arbitrary raster arrays...
        elif mode == 'other':
            if self._subset_id is not None:
                raise NotImplementedError('No way of compositing subsets on arbitrary raster arrays with unknown shape')
            return as_array(file_path, False)[0]
        else:
            raise NotImplementedError('File mode "%s" not recognized' % mode)

    def read_chunked(
            self, file_path, start_row, end_row, mode = None, shp = None):
        '''
        Reads in the file at the given file path, row chunk by row chunk,
        and returns a `numpy.ndarray`.

        Parameters
        ----------
        file_path : str
            The location of a file
        start_row : int
            Index of the first row in a row chunk
        end_row : int
            Index of the final row in a row chunk
        mode : str
            The file mode to use
        shp : tuple
            The shape of the raster arrays

        Returns
        -------
        numpy.ndarray
        '''
        if mode is None:
            mode = self._mode

        r0, r1 = (start_row, end_row)
        if self._subset_id is not None:
            x_idx, y_idx = self.__slice_idx__
            xmin, xmax = x_idx

        # If we're reading HDF5 files...
        if mode == 'hdf5':
            with h5py.File(file_path, 'r') as hdf:
                assert self._field in hdf.keys(), 'Could not find the specified field name: %s' % self._field
                # If we're not subsetting...
                if self._subset_id is None:
                    return hdf[self._field][r0:r1,:]

                return hdf[self._field][r0:r1, xmin:xmax]

        # If we're reading sparse TCF output arrays...
        elif mode in ('sparse', 'tcf'):
            # TCFArray has stronger assumptions than SparseArray; try to read
            #   the file either way
            tcf = SparseArray(file_path, self._grid)

            tcf.inflate()
            if self._subset_id is None:
                return tcf.data[r0:r1,:]

            return tcf.data[r0:r1, xmin:xmax]

        # If we're reading arbitrary raster arrays...
        elif mode == 'other':
            if self._subset_id is None:
                shp = self.__shp__ if shp is None else shp
                return as_array(
                    file_path, False, (0, r0, shp[1], r1))[0]

            return as_array(file_path, False, (xmin, r0, xmax, r1))[0]

        else:
            raise NotImplementedError('File mode "%s" not recognized' % mode)

    def read_raster(self, file_path, mode = None):
        '''
        Reads in the file at the given file path and returns a `gdal.Dataset`.

        Parameters
        ----------
        file_path : str
            The location of a file
        mode : str
            The file mode to use

        Returns
        -------
        gdal.Dataset
        '''
        if mode is None:
            mode = self._mode

        # For all other file modes...
        assert getattr(self, '_grid', None) is not None, 'Must define input --grid in order to read input HDF5 or sparse arrays'
        gt = EASE2_GRID_PARAMS[self._grid]['geotransform']
        wkt = EPSG[EASE2_GRID_PARAMS[self._grid]['epsg']]

        # If we're reading arbitrary raster arrays...
        if mode == 'other':
            rast, _, _ = as_raster(file_path, False)
            return rast

        # If we're reading HDF5 files...
        elif mode == 'hdf5':
            with h5py.File(file_path, 'r') as hdf:
                assert self._field in hdf.keys(), 'Could not find the specified field name: %s' % self._field
                arr = hdf[self._field][:]
                return array_to_raster(arr, gt, wkt)

        # If we're reading sparse TCF output arrays...
        elif mode in ('sparse', 'tcf'):
            # TCFArray has stronger assumptions than SparseArray; try to read
            #   the file either way
            tcf = SparseArray(file_path, self._grid)
            tcf.inflate()
            return array_to_raster(tcf.data, gt, wkt)

        else:
            raise NotImplementedError('File mode "%s" not recognized' % mode)


class ProgressBar(object):
    '''
    An animated progress bar for printing progress to the screen, as with a
    command line interface. Used as a context manager around a loop, e.g.:

        with ProgressBar(len(things), "Working...") as progress:
            for i, each in enumerate(things):
                ...
                progress.update(i)

    Parameters
    ----------
    total : int
        The total number of loop iterations
    prefix : str
        (Optional) The text to print before the progress bar (Default: "")
    suffix : str
        (Optional) The text to print at the end of the progress bar
        (Default: "")
    decimals : int
        (Optional) Number of decimal places for the progress bar's percent
        (Default: 0)
    length : int
        (Optional) The length of the progress bar, in characters (Default: 30)
    fill : str
        (Optional) The character to display for the filled portion of the bar
        (Default: "|")
    verbose : bool
        True to display the bar (Default: True); set to False if the
        `ProgressBar` should not be displayed
    '''
    def __init__(
            self, total, prefix = '', suffix = '', decimals = 0, length = 30,
            fill = '|', verbose = True):
        self._decimals = decimals
        self._fill = fill
        self._length = length
        self._prefix = prefix
        self._suffix = suffix
        self._total = total
        # Some CLIs might have a "verbose" mode, in which progress *should* be
        #   printed to the screen; when not in "verbose," we still enter and
        #   exit the ProgressBar context; so, we return a "dummy" bar that
        #   cannot print to screen when not in "verbose" mode
        self._verbose = verbose

    def __enter__(self):
        self.update(0) # Initialize the bar
        if not self._verbose:
            return self.dummy() # Return a dummy bar
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._verbose:
            self.update(self._total) # Finalize the bar

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, new_prefix):
        self._prefix = new_prefix

    def dummy(self):
        '''
        Returns a dummy instance of ProgressBar that prints nothing; used
        when not in "verbose" mode (the default).
        '''
        dummy = ProgressBar(0)
        setattr(dummy, 'update', lambda s, i: None)
        return dummy

    def update(self, iteration):
        '''
        When called in a loop, creates a progress bar in the terminal.
        NOTE: Adapted from this example:

            https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console

        Parameters
        ----------
        iteration : int
            The current iteration
        total : int
            The total iterations
        prefix : str
            (Optional) Prefix string
        suffix : str
            (Optional) Suffix string
        decimals : int
            (Optional) Positive number of decimals in percent complete
        length : int
            (Optional) Character length of bar
        fill : str
            (Optional) Bar fill character
        '''
        percent = ("{0:." + str(self._decimals) + "f}")\
            .format(100 * (iteration / float(self._total)))
        d = 1 if self._decimals > 0 else 0 # Space for decimal point
        filled_len = int(self._length * iteration // self._total)
        bar = self._fill * filled_len + '-' * (self._length - filled_len)
        # Print new line on complete
        if iteration == self._total:
            # Clear the bar, leave the prefix
            print('\r%s%s' % (self._prefix, ''.rjust(self._length + 10)))
        else:
            print('\r%s [%s] %s%% %s' % (
                self._prefix, bar, percent.rjust(self._decimals + d + 3),
                self._suffix
            ), end = '\r')



'''
Test cases related to the pyl4c.lib.transpiration module.
'''

import os
import unittest
import numpy as np
import pyl4c
from pyl4c.lib.transpiration import canopy_evaporation, psychrometric_constant, radiation_net, svp_slope

class TranspirationTestSuite(unittest.TestCase):
    _depths = -np.array((0.05, 0.1, 0.2, 0.4, 0.75, 1.5)).reshape((6,1))

    def test_psychrometric_constant(self):
        'Should accurately calculate the psychrometric constant'
        pressure = np.array((100e3, 80e3, 100e3, 80e3))
        temp_k = 273.15 + np.array((10, 10, 25, 25))
        answer = [65.74, 52.59, 66.69, 53.35]
        for i in range(0, 4):
            self.assertEqual(
                answer[i],
                psychrometric_constant(pressure[i], temp_k[i]).round(2))
        # Example from FAO:
        #   http://www.fao.org/3/X0490E/x0490e07.htm#psychrometric%20constant%20(g)
        self.assertEqual(
            54.55, np.round(psychrometric_constant(81.8e3, 25 + 273.15), 2))

    def test_radiation_net(self):
        'Should accurately calculate net radiation to the land surface'
        swrad = np.array((500, 5000, 500, 5000, 500, 5000, 500, 5000))
        albedo = np.array((0.4, 0.4, 0.8, 0.8, 0.4, 0.4, 0.8, 0.8))
        temp_k = 273.15 + np.array((10, 10, 10, 10, 25, 25, 25, 25))
        answer = [223.3, 2923.3, 23.3, 923.3, 241.8, 2941.8, 41.8, 941.8]
        for i in range(0, 8):
            self.assertEqual(
                answer[i],
                radiation_net(swrad[i], albedo[i], temp_k[i]).round(1))

    def test_svp_slope(self):
        'Should accurately calculate slope of SVP curve'
        self.assertEqual( 82.3, svp_slope(273.15 + 10).round(1))
        self.assertEqual(144.7, svp_slope(273.15 + 20).round(1))
        self.assertEqual(188.7, svp_slope(273.15 + 25).round(1))

    def test_wet_canopy_evaporation(self):
        'Should accurately calculate wet canopy evaporation'
        evap = canopy_evaporation(
            100e3, 25 + 273.15, 0.5, 1000, 1, 0.5, 5000, 0.1).round(6)
        self.assertEqual(evap, 0.000151)


if __name__ == '__main__':
    unittest.main()



'''
Test cases related to the pyl4c.apps.l4c module.
'''

import datetime
import os
import unittest
import numpy as np
import warnings
import pyl4c
from pyl4c.data.fixtures import BPLUT, restore_bplut
from pyl4c.apps.l4c.main import L4CForwardProcessPoint

INPUTS_FILE_PATH = '/anx_lagr3/arthur.endsley/SMAP_L4C/calibration/v5_Y2020/L4_C_tower_site_drivers_NRv8-3_for_356_sites.h5'
INPUTS_FILE_FOUND = os.path.exists(INPUTS_FILE_PATH)

class L4CForwardProcessPointTestSuite(unittest.TestCase):
    '''
    Integration tests for the L4CForwardProcessPoint model.
    '''
    @classmethod
    def setUpClass(cls):
        cls.default_config = {
            # NOTE: This is the wrong BPLUT version for the data used, but
            #   it doesn't matter because we're just running tests
            'bplut': BPLUT,
            'inputs_file_path': INPUTS_FILE_PATH,
            'site_count': 356,
            'time_steps': 100,
            'start': '2000-01-01',
            'end': (datetime.datetime(2000, 1, 1) + datetime.timedelta(days = 100)).strftime('%Y-%m-%d'),
        }

    @unittest.skipIf(not INPUTS_FILE_FOUND, 'No path to recent input drivers data file')
    def test_model_instantiation(self):
        'Model should be instantiated correctly in debug mode or not'
        config = self.__class__.default_config.copy()
        m0 = L4CForwardProcessPoint(config, verbose = False, debug = True)
        m1 = L4CForwardProcessPoint(config, verbose = False, debug = False)
        for model in (m0, m1):
            self.assertEqual(model.fluxes.shape, (3, 100, 356, 81))
            self.assertEqual(model.fluxes.labels, ('gpp', 'rh', 'nee'))
            self.assertEqual(model.state_initial.shape, (6, 1, 356, 81))
        # If debug = True, should have many more state variables to track
        self.assertEqual(m0.state.shape, (13, 100, 356, 81))
        self.assertEqual(m0.state.labels, (
            'soc1', 'soc2', 'soc3', 'e_mult', 't_mult', 'w_mult', 'f_tmin',
            'f_vpd', 'f_ft', 'f_smrz', 'f_tsoil', 'f_smsf', 'apar'
        ))
        # If debug = False...
        self.assertEqual(m1.state.shape, (6, 100, 356, 81))
        self.assertEqual(m1.state.labels, (
            'soc1', 'soc2', 'soc3', 'e_mult', 't_mult', 'w_mult'
        ))
        self.assertTrue(np.isnan(model.fluxes.data).all())
        self.assertTrue(np.isnan(model.state.data).all())
        self.assertTrue(not np.isnan(model.state_initial.data).all())

    @unittest.skipIf(not INPUTS_FILE_FOUND, 'No path to recent input drivers data file')
    def test_model_instantiation_with_longer_record(self):
        'Model should be instantiated correctly with a longer record'
        config = self.__class__.default_config.copy()
        config['time_steps'] = 150
        config['end'] = (datetime.datetime(2000, 1, 1) +\
            datetime.timedelta(days = 150)).strftime('%Y-%m-%d')
        m1 = L4CForwardProcessPoint(config, verbose = False, debug = False)
        self.assertEqual(m1.state.shape, (6, 150, 356, 81))
        self.assertEqual(m1.fluxes.shape, (3, 150, 356, 81))

    @unittest.skipIf(not INPUTS_FILE_FOUND, 'No path to recent input drivers data file')
    def test_model_constants_setup(self):
        'Model should contain arrays of the parameter values'
        config = self.__class__.default_config.copy()
        m1 = L4CForwardProcessPoint(config, verbose = False, debug = False)
        self.assertEqual(m1.constants.litterfall.shape, (356, 81))
        self.assertEqual(m1.constants.decay_rates.shape, (3, 356, 81))

    @unittest.skipIf(not INPUTS_FILE_FOUND, 'No path to recent input drivers data file')
    def test_model_forward_run(self):
        'Model should have predictable behavior after one forward run step'
        config = self.__class__.default_config.copy()
        m1 = L4CForwardProcessPoint(config, verbose = False, debug = False)
        m1.run(1)
        self.assertTrue(not np.isnan(m1.fluxes.data).all())
        self.assertTrue(np.equal(
            m1.fluxes.data[:,0,0,0].round(3),
            np.array([5.014, 0.939, -2.32]).astype(np.float32)
        ).all())
        self.assertTrue(np.equal(
            m1.fluxes.data[:,0,42,0].round(3),
            np.array([5.765, 2.339, -0.947]).astype(np.float32)
        ).all())
        self.assertTrue(np.equal(
            m1.state.data[:,0,0,0].round(2).astype(np.float64),
            np.array([44.09, 66.11, 2395.2, 0.44, 1, 0.75]).astype(np.float32)
        ).all())
        self.assertTrue(np.equal(
            m1.state.data[:,0,99,0].round(2).astype(np.float64),
            np.array([37.51, 30.29, 2628.11, 0.07, 0.04, 0.46]).astype(np.float32)
        ).all())


if __name__ == '__main__':
    unittest.main()
